"""
SKompiler: Generate Sympy expressions from SKAST.
"""
#pylint: disable=protected-access
import warnings
from collections import OrderedDict
from itertools import product, chain, takewhile, count
import re
import numpy as np
from ._common import ASTProcessor, is_, StandardOps, VectorsAsLists, id_generator


def translate(node, component=None, multistage=True, assign_to=None,
              multistage_subexpression_min_length=3, _max_subexpression_length=8100):
    """Translates SKAST to an Excel formula (or a list of those, if the output should be a vector).

    Kwargs:
        component (int or None):
                   If the result is a vector and only one component is required, pass its index here.

        multistage (bool):
                   When False, generates a single string, describing the model as one long expression.
                   For complex models this string will be too long to be used in Excel.
                   When True (default), returns an ExcelCode object, which is an OrderedDict,
                   mapping cell names to expressions which they should correspond to.

        assign_to: A list or a generator expression, producing Excel cell names
                   which can be filled in a multi-stage computation.
                   When None, a default sequence ['A1', 'B1', ...] will be used.
                   if you would like such sequence, but beginning at, say 'G3',
                   pass excel_range('G3:*3')

        multistage_subexpression_min_length (int):
            Allows to reduce the number of stages in the computation, by preventing the creation
            of an intermediate step whenever the corresponding expression is shorter than the given length.
            I.e. suppose you would like to avoid having a short separate subexpression
                G1 = MAX(A1,B1,C1)
            and would rather have it inlined.
            In this case specify multistage_subexpression_min_length=14 and the expression shorter than 14 characters won't be
            assigned to a separate variable.
            Specifying a very large value is nearly equivalent to setting multistage=False
            (the only difference is that the returned value is still an OrderedDict with a single assignment)

        _max_subexpression_length (int): Max length of a single subexpression in multistage mode.
                                         You should not change it (used for testing internally)

    >>> from skompiler.toskast.string import translate as skast
    >>> expr = skast('[2*x[0]/5, 1] if x[1] <= 3 else [12.0+y, -45.5]')
    >>> print(translate(expr, multistage=False))
    ['IF((x2<=3),((2*x1)/5),(12.0+y))', 'IF((x2<=3),1,(-45.5))']
    >>> print(translate(expr))
    A1=IF((x2<=3),((2*x1)/5),(12.0+y))
    B1=IF((x2<=3),1,(-45.5))
    >>> expr = skast('a=1+x; a+a+a')
    >>> print(translate(expr))
    A1=(1+x)
    B1=((A1+A1)+A1)
    >>> print(translate(expr, multistage=False))
    (((1+x)+(1+x))+(1+x))
    """
    writer = ExcelWriter(multistage=multistage, assign_to=assign_to,
                         multistage_subexpression_min_length=multistage_subexpression_min_length,
                         _max_subexpression_length=_max_subexpression_length)
    result = writer(node)
    if component is not None:
        result = result[component]
    if multistage:
        writer.add_named_subexpression(result)
        return writer.code
    else:
        return result

def _sum(iterable):
    return "({0})".format("+".join(iterable))

def _iif(cond, iftrue, iffalse):
    if iftrue == iffalse:
        return iftrue
    else:
        return 'IF({0},{1},{2})'.format(cond, iftrue, iffalse)

def _dotproduct(xs, ys):
    return _sum('{0}*{1}'.format(x, y) for x, y in zip(xs, ys))

def _step(x):
    return _iif('{0}>0'.format(x), 1, 0)

def _max(xs):
    if len(xs) == 1:
        return xs[0]
    else:
        return 'MAX({0})'.format(','.join(xs))

def _argmax(xs, maxval=None):
    if not maxval:
        maxval = _max(xs)
    expr = str(len(xs)-1)
    n = len(xs)-1
    while n > 0:
        n -= 1
        expr = _iif('{0}={1}'.format(xs[n], maxval), str(n), expr)
    return expr

def is_fmt(template):
    # Auto-compacting binary operator
    def auto_compacting_operator(self, _):
        def fn(x, y=''):
            result = template.format(x, y)
            if self.multistage and len(result) > self.max_subexpression_length:
                if len(x) > len(y):
                    x, y = self.add_named_subexpression(x), y
                else:
                    x, y = x, self.add_named_subexpression(y)
                result = template.format(x, y)
            return result
        return fn
    return auto_compacting_operator

def _takeuntil(value, iterable):
    if value is None:
        return iterable
    else:
        return chain(takewhile(lambda x: x != value, iterable), [value])

_letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

def excel_column_names(start_column='A', end_column=None):
    """
    Generates excel column names starting from a given one.
    end_column is inclusive.

    >>> gen = excel_column_names('C')
    >>> [next(gen) for _ in range(4)]
    ['C', 'D', 'E', 'F']
    >>> list(excel_column_names('Y', 'AC'))
    ['Y', 'Z', 'AA', 'AB', 'AC']
    >>> list(excel_column_names('AY', 'BC'))
    ['AY', 'AZ', 'BA', 'BB', 'BC']
    >>> list(excel_column_names('ZY', 'AAC'))
    ['ZY', 'ZZ', 'AAA', 'AAB', 'AAC']
    """

    ns = _letters
    all_cols = map(''.join, chain(ns, product(ns, ns), product(ns, ns, ns)))
    while next(all_cols) != start_column:
        pass
    return _takeuntil(end_column, chain([start_column], all_cols))

_range_re = re.compile(r'^([\*A-Z]+)(\d+):([\*A-Z]+)(\d+)$')

def excel_range(range_):
    """
    A convenience method for generating lists of excel cells in a row or column.
    The range_ argument is an expression of the form A3:G3
    The second endpoint of the range may contain * instead of the
    row or column coordinate (as in A3:*3 or A3:A*), in this case
    the return value is a generator, enumerating cells in the given
    row or column indefinitely.
    """
    matches = _range_re.match(range_.upper())
    if not matches:
        raise ValueError("Range must be of the form A1:B1")
    lcol, lrow, rcol, rrow = matches.groups()
    if lcol == '*' or lrow == '*':
        raise ValueError("Only the right side of the interval may include wildcard")
    if lcol == rcol:  # Fixed column
        if rrow == '*':
            rows = count(int(lrow))
        else:
            rows = range(int(lrow), int(rrow)+1)
        yield from ('{0}{1}'.format(lcol, i) for i in rows)
    elif lrow == rrow: # Fixed row
        cols = excel_column_names(lcol, None if rcol == '*' else rcol)
        yield from ('{0}{1}'.format(col, lrow) for col in cols)
    else:
        raise ValueError("Only single-column or single-row ranges are supported")

def _compact_string(s, max_len=70):
    if len(s) > max_len:
        part = (max_len - 30) // 2
        return s[:part] + ' ...{0} chars skipped... '.format(len(s)-2*part) + s[-part:]
    else:
        return s

_builtins = {
    'IF': lambda t, a, b: a if t else b,
    'MAX': max,
    'EXP': np.exp,
    'LOG': np.log,
    'SQRT': np.sqrt,
    'ABS': np.abs,
}
_single_comparison = re.compile(r'(?<!\<)=')

class ExcelCode(OrderedDict):
    """A version of OrderedDict which prints itself nicer."""
    def __str__(self):
        lines = ['{0}={1}'.format(k, _compact_string(v))
                 for k, v in self.items()]
        if len(lines) > 10:
            lines = lines[:4] + ['      ... {0} lines skipped ...'.format(len(lines)-8)] +\
                    lines[-4:]
        return '\n'.join(lines)

    def to_dataframe(self):
        """Converts code to a pandas dataframe,
        suitable for pasting into Excel.

        The main usecase for this method is:
            
            code.to_dataframe().to_clipboard()

        """
        import pandas as pd
        return pd.DataFrame([[f'={v}' for v in self.values()]],
                            columns=self.keys())

    def evaluate(self, **kwargs):
        """Evaluates the excel code using Python's eval.
        Will probably fail with MemoryError for longer strings
        (because Python's ast.parse can't handle them)."""
        env = {}
        env.update(_builtins)
        for k, v in self.items():
            expand = _single_comparison.sub('==', v) # Excel uses '=' for comparisons
            env[k] = eval(expand, kwargs, env) #pylint: disable=eval-used
        for k in _builtins:
            del env[k]
        return env


class ExcelWriter(ASTProcessor, StandardOps, VectorsAsLists):
    """A SK AST processor, producing an Excel expression (or a list of those)"""

    def __init__(self, multistage=False, assign_to=None, positive_infinity=float(np.finfo('float64').max),
                 negative_infinity=float(np.finfo('float64').min),
                 multistage_subexpression_min_length=3,
                 _max_subexpression_length=8100):
        self.positive_infinity = positive_infinity
        self.negative_infinity = negative_infinity
        self.multistage = multistage
        self.multistage_subexpression_min_length = multistage_subexpression_min_length
        self.max_subexpression_length = _max_subexpression_length  # In multistage mode we attempt to keep subexpressions shorter than this
                                                                   # (because Excel does not allow cell values longer than 8196 chars)
                                                                   # NB: this is rather ad-hoc and may not always work.

        if self.multistage:
            if assign_to is None:
                warnings.warn("Value of the assign_to parameter is not provided. Will use default ['A1', 'B1', ...']", UserWarning)
                assign_to = excel_range('A1:*1')
            self.assign_to = assign_to if hasattr(assign_to, '__next__') else iter(assign_to)
            self.code = ExcelCode()
            self.references = [{}]
            self.temp_ids = id_generator()

    def Identifier(self, id):
        return id.id

    def IndexedIdentifier(self, sub):
        warnings.warn("Excel does not support vector types natively. "
                      "Numbers will be appended to the given feature name, "
                      "it may not be what you intend.", UserWarning)
        return "{0}{1}".format(sub.id, sub.index+1)
    
    def NumberConstant(self, num):
        # Infinities have to be handled separately
        if np.isinf(num.value):
            val = self.positive_infinity if num.value > 0 else self.negative_infinity
        else:
            val = num.value
        return str(val)

    def _iif(self, test, ift, iff):
        # Auto-compacting IIF
        result = _iif(test, ift, iff)
        if self.multistage and len(result) > self.max_subexpression_length:
            if len(ift) > len(iff):
                ift, iff = self.add_named_subexpression(ift), iff
            else:
                ift, iff = ift, self.add_named_subexpression(iff)
            result = _iif(test, ift, iff)
        return result
        
    # Implement binary and unary operations
    Mul = is_fmt('({0}*{1})')
    Div = is_fmt('({0}/{1})')
    Add = is_fmt('({0}+{1})')
    Sub = is_fmt('({0}-{1})')
    LtEq = is_fmt('({0}<={1})')
    Eq = is_fmt('({0}={1})')
    USub = is_fmt('(-{0})')
    Exp = is_fmt('EXP({0})')
    Sqrt = is_fmt('SQRT({0})')
    Log = is_fmt('LOG({0})')
    Max = is_fmt('MAX({0},{1})')
    Abs = is_fmt('ABS({0})')
    Step = is_(_step)
    Sigmoid = is_fmt('(1/(1+EXP(-{0})))')
    DotProduct = is_(_dotproduct)
    VecSum = is_(_sum)
    VecMax = is_(_max)
    MatVecProduct = lambda self, _: self._matvecproduct
    ArgMax = lambda self, _: self._argmax
    Softmax = lambda self, _: self._softmax

    def _argmax(self, xs):
        xs = [self.possibly_add_named_subexpression(x) for x in xs]
        max_var = self.possibly_add_named_subexpression(_max(xs))
        return _argmax(xs, max_var)

    def _matvecproduct(self, M, xs):
        xs = [self.possibly_add_named_subexpression(x) for x in xs]
        return [_sum('{0}*{1}'.format(m_i[j], xs[j]) for j in range(len(xs))) for m_i in M]
    
    def _vecsumnormalize(self, xs):
        xs = [self.possibly_add_named_subexpression(x) for x in xs]
        sum_var = self.possibly_add_named_subexpression(_sum(xs))
        return ['({0}/{1})'.format(x, sum_var) for x in xs]

    def _softmax(self, xs):
        return self._vecsumnormalize(['EXP({0})'.format(x) for x in xs])

    def Let(self, node, **kw):
        if not self.multistage:
            return StandardOps.Let(self, node)
        else:
            self.references.append({})
            for defn in node.defs:
                self.add_named_subexpression(self(defn.body), defn.name)
            result = self(node.body)
            self.references.pop()
            return result

    def Reference(self, node):
        if not self.multistage:
            raise ValueError("Reference nodes are only supported in multi-stage code generation")
        if node.name not in self.references[-1]:
            raise ValueError("Undefined reference: {0}".format(node.name))
        return self.references[-1][node.name]
    
    def possibly_add_named_subexpression(self, value):
        if self.multistage and len(value) >= self.multistage_subexpression_min_length:
            return self.add_named_subexpression(value)
        else:
            return value
        
    def add_named_subexpression(self, value, name=None):
        is_list = isinstance(value, list)
        if not isinstance(value, list):
            value = [value]
        if name is None:
            name = next(self.temp_ids)
        try:
            ref = []
            for v in value:
                next_output = next(self.assign_to)
                if next_output in self.code:
                    raise ValueError("Repeated names are not supported in the assign_to parameter")
                self.code[next_output] = v
                ref.append(next_output)
            if len(ref) == 1 and not is_list:
                ref = ref[0]
            self.references[-1][name] = ref
            return ref
        except StopIteration as ex:
            raise ValueError("The number of fields provided in the assign_to parameter"
                             " is not sufficient to complete the computation.") from ex



"""
AdaBoost
"""
import numpy as np
from skompiler.dsl import const, func, sum_
from ..common import classifier, vecsumnormalize
from ..tree.base import decision_tree

# NB: AdaboostRegressor is annoying to implement as it requires
# finding a weighted median among the all estimator predictions,
# which is not meaningfully implementable without adding extra special functions
# Thus we only support AdaboostClassifier for now
def adaboost_classifier(model, inputs, method="predict_proba"):
    """
    Creates a SKAST expression corresponding to a given adaboost classifier.
    """
    divisor = model.estimator_weights_.sum()
    if method == 'decision_function':
        divisor /= (model.n_classes_ - 1)
    tree_exprs = [decision_tree(e.tree_,
                                method='predict_proba' if (model.algorithm == 'SAMME.R' or
                                                           method.startswith('predict_')) else 'predict',
                                inputs=inputs,
                                value_transform=adaboost_values(model, w/divisor, method))
                  for e, w in zip(model.estimators_, model.estimator_weights_)]
    decision = sum_(tree_exprs)

    if method == 'decision_function':
        if model.n_classes_ == 2:
            decision = decision @ const([-1, 1])
        return decision
    elif method == 'predict':
        return func.ArgMax(decision)
    else:
        return classifier(vecsumnormalize(func.Exp(sum_(tree_exprs)), model.n_classes_), method)


def adaboost_values(m, weight=1.0, method='predict_proba'):
    def _samme(proba):
        proba = np.array(proba)
        proba[proba < np.finfo(proba.dtype).eps] = np.finfo(proba.dtype).eps
        log_proba = np.log(proba)
        return (log_proba - (1. / m.n_classes_) * log_proba.sum(axis=1)[:, np.newaxis])*weight
    
    def _ada_predict_proba(probs):
        return probs * weight / (m.n_classes_ - 1)
    
    def _ada_predict(preds):
        probs = np.zeros((len(preds), m.n_classes_))
        probs[np.arange(len(preds)), preds] = 1
        probs *= weight/(m.n_classes_-1)
        return probs

    if m.algorithm == 'SAMME.R':
        return _samme
    elif method == 'predict' or method == 'decision_function':
        return _ada_predict
    else:
        return _ada_predict_proba



"""
SKompiler: AST nodes.

The classes here describe the AST nodes of the expressions produced
by SKompiler.

Notes:
    - We might have relied on Python's ast.* classes, but this
      would introduce unnecesary complexity and limit possibilities for
      adding custom nodes for special cases.
    - @dataclass would be a nice tech to use here, but it would prevent
      compatibility with Python older than 3.6

>>> expr = BinOp(Mul(), Identifier('x'), IndexedIdentifier('y', -5, 10))
>>> expr = BinOp(Add(), NumberConstant(12.2), expr)
>>> print(str(expr))
(12.2 + (x * y[-5]))
>>> expr = Let([Definition('z', expr)], BinOp(Add(), Reference('z'), NumberConstant(2)))
>>> print(str(expr))
{
$z = (12.2 + (x * y[-5]));
($z + 2)
}
"""
#pylint: disable=protected-access,multiple-statements,too-few-public-methods,no-member
from itertools import count
from importlib import import_module
import numpy as np


# Each ASTNode registers its class name in this set
AST_NODES = set([])

# This is the set of conversions supported in the node.to(...) method
TRANSLATORS = {
    'excel': 'skompiler.fromskast.excel:translate',
    'python': 'skompiler.fromskast.python:translate',
    'sqlalchemy': 'skompiler.fromskast.sqlalchemy:translate',
    'sympy': 'skompiler.fromskast.sympy:translate',
    'string': str
}


#region ASTNode base -------------------------------------------------

# Basic type inference (NB: not an enum to keep compatibility with Python 3.3. Maybe it is lost already, though)
DTYPE_SCALAR = 1
DTYPE_VECTOR = 2
DTYPE_MATRIX = 3
DTYPE_OTHER = 42
DTYPE_UNKNOWN = None
class ASTTypeError(ValueError): pass
class UnableToDecompose(ValueError): pass


class ASTNodeCreator(type):
    """A metaclass, which allows us to implement our AST nodes like mutable namedtuples
       with a nicer declaration syntax."""
    def __new__(mcs, name, bases, dct, fields='', repr=None, dtype=DTYPE_OTHER):
        if fields is None:
            return super().__new__(mcs, name, bases, dct)
        else:
            cls = super().__new__(mcs, name, bases, dct)
            cls._fields = fields.split()
            cls._template = repr
            cls._default_dtype = dtype
            AST_NODES.add(name)
            return cls

    # For Python 3.5, see https://stackoverflow.com/a/25191150/318964
    def __init__(cls, name, bases, dct, **_):
        super().__init__(name, bases, dct)


_singletons = {}

class ASTNode(object, metaclass=ASTNodeCreator, fields=None):
    """Base class for all AST nodes. You may not instantiate it."""

    def __new__(cls, *_args, **_kw):
        # Save some memory on singletons
        if hasattr(cls, '_fields') and not cls._fields:
            # Singleton node
            name = cls.__name__
            if _singletons.get(name, None) is None:
                _singletons[name] = super().__new__(cls)
            return _singletons[name]
        return super().__new__(cls)

    def __init__(self, *args, **kw):
        """Make sure the constructor arguments correspond to the _fields proprty"""
        
        if len(args) > len(self._fields): raise Exception("Too many arguments")
        args = dict(zip(self._fields, args))
        for k in args:
            if k in kw:
                raise Exception("Argument %s defined multiple times" % k)
        args.update(kw)
        if len(args) != len(self._fields):
            raise Exception("Not enough arguments ({0}) given to {1}".format(len(args), self.__class__.__name__))
        for k in self._fields:
            if k not in args:
                raise Exception("Argument %s not provided" % k)
            self.__dict__[k] = args[k]
        self.__dict__['_dtype'] = self._compute_dtype()
    
    def __str__(self):
        dct = {k: str(v) for k, v in vars(self).items()}
        return self._template.format(**dct)

    def __repr__(self):
        return self.__class__.__name__ + '(' + ', '.join(k + '=' + repr(getattr(self, k)) for k in self._fields) + ')'

    def __setattr__(self, field, value):
        raise Exception("AST nodes are immutable")

    def __iter__(self):
        "Iteration over node fields. Note that the node works like a dict.items(), not like a tuple"
        for k in self._fields:
            yield k, getattr(self, k)

    def __bool__(self):
        return True
    
    # Convenience routines for combining AST nodes
    def __len__(self):
        raise UnableToDecompose()
    
    def __getitem__(self, idx):
        raise UnableToDecompose()

    def __add__(self, other):
        other = self._align_scalar(other)
        return BinOp(Add(), self, other)
    
    def __mul__(self, other):
        other = self._align_scalar(other)
        return BinOp(Mul(), self, other)
    
    def __truediv__(self, other):
        other = self._align_scalar(other)
        return BinOp(Div(), self, other)
    
    def __sub__(self, other):
        other = self._align_scalar(other)
        return BinOp(Sub(), self, other)
    
    def __matmul__(self, other):
        if self._dtype == DTYPE_MATRIX:
            return BinOp(MatVecProduct(), self, other)
        else:
            return BinOp(DotProduct(), self, other)

    def __le__(self, other):
        other = self._align_scalar(other)
        return BinOp(LtEq(), self, other)
    
    def __eq__(self, other):
        other = self._align_scalar(other)
        return BinOp(Eq(), self, other)
    
    def __call__(self, arg, arg2=None):
        if arg2 is not None:
            return BinOp(self, arg, arg2)
        else:
            return UnaryFunc(self, arg)
    
    def _align_scalar(self, other):
        # If a scalar is given in a binary operation,
        # we try to align it in length with us
        # This is hackish and should be used with caution
        if np.isscalar(other):
            self_type = self._dtype
            if self_type == DTYPE_SCALAR:
                return NumberConstant(other)
            elif self_type == DTYPE_VECTOR:
                return VectorConstant([other]*len(self))
            else:
                raise ASTTypeError("Unable to align scalar with a node of type {0}".format(self.__class__.__name__))
        else:
            return other

    # Type inference
    def _compute_dtype(self):
        return self._default_dtype

    # Convenience routines for evaluating AST nodes
    def lambdify(self):
        """
        Converts the SKAST expression to an executable Python function.

        >>> from .toskast.string import translate as skast
        >>> skast("12.4 * (X[1] + Y)").lambdify()(**{'X': [10, 20, 30], 'Y': 1.2})
        262.88
        >>> skast("122.45 + 1").lambdify()()
        123.45
        >>> skast("x[0]").lambdify()(x=[[1]])
        [1]
        >>> skast("a=1; a").lambdify()()
        1
        >>> skast("a=b; b=b+1; c=b+b; a+b+c").lambdify()(b=1.0)
        7.0
        """
        from .fromskast import python

        return python.lambdify(python.translate(self))

    def evaluate(self, **inputs):
        "Convenience routine for evaluating expressions"
        return self.lambdify()(**inputs)

    # The main convenience function, which converts the node to any of the
    # supported formats
    def to(self, target, *args, **kw):
        """Convenience routine for converting expressions to any
           supported output form.
           See project documentation for detailed explanation.

           Equivalent to skompiler.fromskast.<dialect[0]>.translate(self, dialect[1], *args, **kw)
           (where dialect[0] and dialect[1] denote the parts to the left and right of the '/' in the
            dialect parameter)

         Args:
         
            target (str): The target value. Possible values are:
            
               - 'sqlalchemy',
               - 'sqlalchemy/<dialect>', where <dialect> is on of the supported
                 SQLAlchemy dialects ('firebird', 'mssql', 'mysql', 'oracle',
                 'postgresql', 'sqlite', 'sybase')
               - 'sympy'
               - 'sympy/<lang>', where <lang> is either of:
                 'c', 'cxx', 'rust', 'fortran', 'js', 'r', 'julia',
                 'mathematica', 'octave'
               - 'excel'
               - 'python'
               - 'python/code'
               - 'python/lambda'
               - 'string'
            
        Kwargs:

            *args, **kw: Extra arguments are passed to the translation function.
                         
                The most important for `sqlalchemy` and `sympy/<lang>` dialects are:
                
                   assign_to:  When not None, the generated code outputs the result to a variable
                          (or variables, or columns) with given name(s).
                   component:  When not None and the expression produces a vector, only the
                            specified component of the vector is output (0-indexed)
                            
                For more info, see documentation for
                    skompiler.fromskast.<dialect[0]>.translate.

        '"""
        translator, *dialect = target.split('/')
        if translator not in TRANSLATORS:
            raise ValueError("Invalid translator: {0}".format(translator))
        translator = TRANSLATORS[translator]
        if not hasattr(translator, '__call__'):
            module, callable = translator.split(':')
            mod = import_module(module)
            translator = getattr(mod, callable)
        return translator(self, *dialect, *args, **kw)
#endregion


#region AST node types ---------------------------------------------

# Unary operators and functions
class UnaryFunc(ASTNode, fields='op arg', repr='{op}({arg})'):
    def __len__(self):
        if isinstance(self.op, IsElemwise):
            return len(self.arg)
        else:
            raise UnableToDecompose()
    
    def __getitem__(self, index):
        if isinstance(self.op, IsElemwise):
            return UnaryFunc(self.op, self.arg[index])
        else:
            raise UnableToDecompose()

    def _compute_dtype(self):
        if self.arg._dtype not in [DTYPE_UNKNOWN, DTYPE_SCALAR, DTYPE_VECTOR]:
            raise ASTTypeError()
        if isinstance(self.op, IsElemwise):
            return self.arg._dtype
        else:
            return self.op._out_dtype

# Some unary functions distribute over vectors. We mark them as such
class IsElemwise: pass
class USub(ASTNode, IsElemwise, repr='-'): pass     # Unary minus operator
class Exp(ASTNode, IsElemwise, repr='exp'): pass
class Log(ASTNode, IsElemwise, repr='log'): pass
class Abs(ASTNode, IsElemwise, repr='abs'): pass
class Sqrt(ASTNode, IsElemwise, repr='sqrt'): pass
class Step(ASTNode, IsElemwise, repr='step'): pass  # Heaviside step (x > 0)
class Sigmoid(ASTNode, IsElemwise, repr='sigmoid'): pass

# Some functions take vector arguments but do not distribute elementwise
class VecSum(ASTNode, repr='sum'):
    _out_dtype = DTYPE_SCALAR
class VecMax(ASTNode, repr='max'):
    _out_dtype = DTYPE_SCALAR
class ArgMax(ASTNode, repr='argmax'):
    _out_dtype = DTYPE_SCALAR
class Softmax(ASTNode, repr='softmax'):
    _out_dtype = DTYPE_VECTOR

# Binary operators
def _common_dtype(nodes):
    dtypes = {n._dtype for n in nodes if n._dtype is not DTYPE_UNKNOWN}
    if not dtypes:
        return DTYPE_UNKNOWN
    elif len(dtypes) == 2:
        raise ASTTypeError("Mismatching operand types")
    else:
        result = next(iter(dtypes))
        if result not in [DTYPE_SCALAR, DTYPE_VECTOR]:
            raise ASTTypeError("Arguments must be scalars or vectors")
        return result


class BinOp(ASTNode, fields='op left right', repr='({left} {op} {right})'):
    def __len__(self):
        if isinstance(self.op, IsElemwise) or isinstance(self.op, MatVecProduct):
            return len(self.left)
        else:
            raise UnableToDecompose()
    
    def __getitem__(self, index):
        if isinstance(self.op, IsElemwise):
            return BinOp(self.op, self.left[index], self.right[index])
        elif isinstance(self.op, MatVecProduct):
            return BinOp(DotProduct(), self.left[index], self.right)
        else:
            raise UnableToDecompose()

    def _compute_dtype(self):
        if isinstance(self.op, IsElemwise):
            return _common_dtype([self.left, self.right])
        else:
            return self.op._out_dtype

class Mul(ASTNode, IsElemwise, repr='*'): pass
class Add(ASTNode, IsElemwise, repr='+'): pass
class Sub(ASTNode, IsElemwise, repr='-'): pass
class Div(ASTNode, IsElemwise, repr='/'): pass
class Max(ASTNode, IsElemwise, repr='max'): pass
class DotProduct(ASTNode, repr='v@v'):
    _out_dtype = DTYPE_SCALAR
class MatVecProduct(ASTNode, repr='m@v'):
    _out_dtype = DTYPE_VECTOR

class LFold(ASTNode, fields='op elems'):
    """
    Left-associative fold of an operator over a list of arguments
    E.g. sum(xs) := LFold(Add(), xs)
    This could be represented as a sequence of binary ops, but having a
    dedicated operator may significantly reduce the depth of AST trees with long sums
    which you may find in ensemble classifiers.
    Deep trees are bad because you run the risk of hitting system recursion limit when processing
    them via recursive parsers (as is the case currently)
    """

    def __str__(self):
        return str(self.op).join(str(e) for e in self.elems)

    def __getitem__(self, idx):
        return LFold(self.op, [el[idx] for el in self.elems])

    def __len__(self):
        if not self.elems:
            raise ASTTypeError("Empty LFolds are not allowed")
        return len(self.elems[0])

    def _compute_dtype(self):
        return _common_dtype(self.elems)

# Boolean binary ops
class IsBoolean: pass
class LtEq(ASTNode, IsElemwise, IsBoolean, repr='<='): pass
class Eq(ASTNode, IsElemwise, IsBoolean, repr='=='): pass


# IfThenElse
class IfThenElse(ASTNode, fields='test iftrue iffalse', repr='(if {test} then {iftrue} else {iffalse})'):
    def __len__(self):
        return len(self.iftrue)
    
    def __getitem__(self, index):
        return IfThenElse(self.test, self.iftrue[index], self.iffalse[index])

    def _compute_dtype(self):
        return _common_dtype([self.iftrue, self.iffalse])

# Special function
class MakeVector(ASTNode, fields='elems', dtype=DTYPE_VECTOR):
    def __str__(self):
        elems = ', '.join(str(e) for e in self.elems)
        return '[{0}]'.format(elems)

    def __getitem__(self, idx):
        return self.elems[idx]

    def __len__(self):
        return len(self.elems)

# Leaf nodes
class IsAtom: pass
class VectorIdentifier(ASTNode, IsAtom, fields='id size', repr='{id}', dtype=DTYPE_VECTOR):
    def __getitem__(self, index):
        return IndexedIdentifier(self.id, index, self.size)
    def __len__(self):
        return self.size

class Identifier(ASTNode, IsAtom, fields='id', repr='{id}', dtype=DTYPE_SCALAR): pass

# Note that IndexedIdentifier is not a generic subscript operator. Its field must contain a string id and an integer index as well as the
# total size of the vector being indexed.
# This lets us "fake" vector input variables in contexts like SQL, where we interpret IndexedIdentifier("x", 1, 10) as a concatenated name "x1"
class IndexedIdentifier(ASTNode, IsAtom, fields='id index size', repr='{id}[{index}]', dtype=DTYPE_SCALAR): pass
class NumberConstant(ASTNode, IsAtom, fields='value', repr='{value}', dtype=DTYPE_SCALAR): pass
class VectorConstant(ASTNode, IsAtom, fields='value', repr='{value}', dtype=DTYPE_VECTOR):
    def __getitem__(self, index):
        return NumberConstant(self.value[index])
    def __len__(self):
        return len(self.value)

class MatrixConstant(ASTNode, IsAtom, fields='value', repr='{value}', dtype=DTYPE_MATRIX):
    def __len__(self):
        return len(self.value)
    def __getitem__(self, index):
        return VectorConstant(self.value[index])

# Variable definitions
# NB: at the moment let-scopes do not capture the outside variables.
# I.e. all the references inside the definitions and body of a let expressions are assumed to
# refer to variables defined in this let scope
class Let(ASTNode, fields='defs body'):
    def __str__(self):
        defs = ';\n'.join(str(d) for d in self.defs)
        return '{\n' + defs + ';\n' + str(self.body) + '\n}'
    def __len__(self):
        return len(self.body)
    def __getitem__(self, idx):
        return self.body[idx]
    def _compute_dtype(self):
        return self.body._dtype

class Definition(ASTNode, fields='name body', repr='${name} = {body}'):
    def _compute_dtype(self):
        return self.body._dtype

class Reference(ASTNode, fields='name', repr='${name}', dtype=DTYPE_UNKNOWN): pass

# Sometimes marking the type and dimension of the object referenced to is convenient
class TypedReference(ASTNode, fields='name dtype size', repr='${name}'):
    def _compute_dtype(self):
        return self.dtype
    def __len__(self):
        return self.size

#endregion


#region Node processing functions -----------------------
def map_list(node_list, fn, call_on_enter=False, call_on_exit=True):
    new_list = []
    changed = False
    for node in node_list:
        new_node = map_tree(node, fn, call_on_enter, call_on_exit)
        if new_node is not node:
            changed = True
        new_list.append(new_node)
    return new_list if changed else node_list


def map_tree(node, fn, call_on_enter=False, call_on_exit=True):
    """Applies a function to each node in the tree.

    Args:
        fn: function, which must accept (node, is_entering) and return a node.
    
    Kwargs:
        call_on_enter:  Invoke the function (potentially replacing the node) every time before entering the node.
        call_on_exit:   Invoke the function (potentially replacing the node) every time before leavin the node.
    """
    if call_on_enter: node = fn(node, True)

    updates = {}
    for fname, fval in node:
        if isinstance(fval, ASTNode):
            new_val = map_tree(fval, fn, call_on_enter, call_on_exit)
        elif fname in ['elems', 'defs']:  # Special case for MakeVector and Let nodes
            new_val = map_list(fval, fn, call_on_enter, call_on_exit)
        else:
            new_val = fval
        if new_val is not fval:
            updates[fname] = new_val
    node = replace(node, **updates)

    if call_on_exit: node = fn(node, False)
    return node


def substitute_references(node, definitions):
    """Substitutes all references in the given expression with ones from the `definitions` dictionary.
       If any substitutions were made, returns a new node. Otherwise returns the same node.

       If references with no matches are found, raises a ValueError.

    Args:
       definitions (dict): a dictionary (name -> expr)

    >>> expr = BinOp(Add(), Identifier('x'), NumberConstant(2))
    >>> substitute_references(expr, {}) is expr
    True
    >>> expr = replace(expr, left = Reference('x'))
    >>> substitute_references(expr, {})
    Traceback (most recent call last):
    ...
    ValueError: Unknown variable reference: x
    >>> new_expr = substitute_references(expr, {'x': expr})
    >>> print(new_expr)  # Only one level of references is expanded
    (($x + 2) + 2)
    >>> assert new_expr is not expr
    >>> assert new_expr.right is expr.right
    """
    def fn(node, _):
        if isinstance(node, Reference) or isinstance(node, TypedReference):
            if node.name not in definitions:
                raise ValueError("Unknown variable reference: " + node.name)
            else:
                return definitions[node.name]
        else:
            return node

    return map_tree(node, fn, call_on_enter=False, call_on_exit=True)


def inline_definitions(let_expr):
    """Given a Let expression, substitutes all the definitions and returns a single
       evaluatable non-let expression.

    >>> from .toskast.string import translate as skast
    >>> expr = inline_definitions(skast('a=1; a'))
    >>> print(str(expr))
    1
    >>> expr = inline_definitions(skast('a=1; b=a+a+2; c=b+b+3; a+b+c+X'))
    >>> print(str(expr))
    (((1 + ((1 + 1) + 2)) + ((((1 + 1) + 2) + ((1 + 1) + 2)) + 3)) + X)
    >>> expr = inline_definitions(skast('X+Y'))
    Traceback (most recent call last):
    ...
    ValueError: Let expression expected
    """
    if not isinstance(let_expr, Let):
        raise ValueError("Let expression expected")
    
    let_expr = merge_let_scopes(let_expr)
    
    defs = {}
    for defn in let_expr.defs:
        defs[defn.name] = substitute_references(defn.body, defs)
    
    return substitute_references(let_expr.body, defs)


# Let expression extraction
def _scope_id_gen():
    yield ''  # Do not mangle the scope of the first let we find
    yield from map('_{0}'.format, count())

class LetCollector:
    def __init__(self):
        self.definitions = []
        self.scopes = []
        self.temp_ids = _scope_id_gen()

    def __call__(self, node, entering):
        if isinstance(node, Let):
            if entering:
                name = next(self.temp_ids)
                self.scopes.append(name)
            else:
                self.scopes.pop()
                return node.body
        elif isinstance(node, Definition) and not entering:
            self.definitions.append(replace(node, name='{0}{1}'.format(self.scopes[-1], node.name)))
        elif isinstance(node, Reference) and not entering:
            if not self.scopes:
                raise ValueError("Undefined reference: {0}".format(node.name))
            return replace(node, name='{0}{1}'.format(self.scopes[-1], node.name))
        return node


def merge_let_scopes(expr):
    """
    Given an expression which may potentially contain Let subexpressions inside,
    bubbles them all up and returns a single Let expression.
    If the original expression contained no Let subexpressions, returns it as-is.

    >>> from skompiler.toskast.string import translate as skast
    >>> let1 = skast("x=1; y=2; x+y")   # 3
    >>> let2 = skast("x=3; y=4; x+2*y") # 11
    >>> let3 = skast("x=0; y=0; 3*x+y") # 20
    >>> let3.defs[0].__dict__['body']=let1
    >>> let3.defs[1].__dict__['body']=let2
    >>> merged = merge_let_scopes(let3)
    >>> print(merged)
    {
    $_0x = 1;
    $_0y = 2;
    $x = ($_0x + $_0y);
    $_1x = 3;
    $_1y = 4;
    $y = ($_1x + (2 * $_1y));
    ((3 * $x) + $y)
    }
    >>> print(inline_definitions(merged))
    ((3 * (1 + 2)) + (3 + (2 * 4)))
    """
    lc = LetCollector()
    expr = map_tree(expr, lc, True, True)
    if not lc.definitions:
        return expr
    else:
        return Let(lc.definitions, expr)


# Node copying and field replacing methods
# NB: We don't implement them as methods to avoid potential conflicts with
# field names ("lambdify", "evaluate" and "to" are the only exceptions as they might be used more often)

def copy(node):
    """
    Copies the node.

    >>> o = BinOp(Add(), NumberConstant(2), NumberConstant(3))
    >>> o2 = copy(o)
    >>> o.__dict__['op'] = Sub()
    >>> print(o, o2)
    (2 - 3) (2 + 3)
    """
    return node.__class__(**dict(node))

def replace(node, **kw):
    """Equivalent to namedtuple's _replace.
    When **kw is empty simply returns node itself.
    
    >>> o = BinOp(Add(), NumberConstant(2), NumberConstant(3))
    >>> o2 = replace(o, op=Sub(), left=NumberConstant(3))
    >>> print(o, o2)
    (2 + 3) (3 - 3)
    """
    if kw:
        vals = dict(node)
        vals.update(**kw)
        return node.__class__(**vals)
    else:
        return node

def decompose(node):
    return [node[i] for i in range(len(node))]

#endregion



"""
Base class for AST processors and functions, useful within multiple implementations
"""
#pylint: disable=not-callable
from itertools import count
from functools import reduce
from ..ast import AST_NODES, inline_definitions, IndexedIdentifier, NumberConstant, Exp, BinOp, IsElemwise, Reference


class ASTProcessorMeta(type):
    """A metaclass, which checks that the class defines methods for all known AST nodes.
       This is useful to verify SKAST processor implementations for completeness."""
    
    def __new__(mcs, name, bases, dct):
        if name != 'ASTProcessor':
            # This way the verification applies to all subclasses of ASTProcessor
            unimplemented = AST_NODES.difference(dct.keys())
            # Maybe the methods are implemented in one of the base classes?
            for base_cls in bases:
                unimplemented.difference_update(dir(base_cls))
            if unimplemented:
                raise ValueError(("Class {0} does not implement all the required ASTParser methods. "
                                  "Unimplemented methods: {1}").format(name, ', '.join(unimplemented)))
        return super().__new__(mcs, name, bases, dct)


class ASTProcessor(object, metaclass=ASTProcessorMeta):
    """
    The class hides the need to specify ASTProcessorMeta metaclass
    """

    def __call__(self, node, **kw):
        return getattr(self, node.__class__.__name__)(node, **kw)


def is_(val):
    return lambda self, node: val

def tolist(x):
    if hasattr(x, 'tolist'):
        return x.tolist()
    else:
        return list(x)

def not_implemented(self, node, *args, **kw):
    raise NotImplementedError("Processing of node {0} is not implemented.".format(node.__class__.__name__))

def _apply_bin_op(op_node, op, left, right):
    if (not isinstance(left, list) and not isinstance(right, list)) or not isinstance(op_node, IsElemwise):
        return op(left, right)
    if not isinstance(left, list) or not isinstance(right, list):
        raise ValueError("Elementwise operations requires both operands to be lists")
    if len(left) != len(right):
        raise ValueError("Sizes of the arguments do not match")
    return [op(l, r) for l, r in zip(left, right)]


class StandardOps:
    """Common implementation for BinOp, UnaryFunc, LFold, Let and ArgMin"""

    def BinOp(self, node, **kw):
        """Most common implementation for BinOp,
        If the arguments are lists and the op is elemwise, applies
        the operation elementwise and returns a list."""

        left = self(node.left, **kw)
        right = self(node.right, **kw)
        op = self(node.op, **kw)
        return _apply_bin_op(node.op, op, left, right)

    def UnaryFunc(self, node, **kw):
        op, arg = self(node.op, **kw), self(node.arg, **kw)
        if not isinstance(node.op, IsElemwise) or not isinstance(arg, list):
            return op(arg)
        else:
            return [op(a) for a in arg]
    
    def LFold(self, node, **kw):
        # Standard implementation simply expands LFold into a sequence of BinOps and then calls itself
        if not node.elems:
            raise ValueError("LFold expects at least one element")
        return self(reduce(lambda x, y: BinOp(node.op, x, y), node.elems), **kw)
    
    def Let(self, node, **kw):
        "Lazy implementation of the 'Let' node. Simply substitutes variables and proceeds as normal."
        return self(inline_definitions(node), **kw)

    Reference = Definition = not_implemented

    def TypedReference(self, node, **_):
        return self(Reference(node.name))


class VectorsAsLists:
    """A partial implementation of an AST processor,
    which assumes that:
        - all vectors are implemented as lists,
        - all element-wise operations operate element-wise on the lists,
        - all binary and unary operations are interpreted to lambda functions.

    The impementation of IfThenElse requires the class to have an
    _iif function, which corresponds to a unary IfThenElse.
    """

    _iif = not_implemented   # Must implement this method in subclasses

    def VectorIdentifier(self, id):
        return [self(IndexedIdentifier(id.id, i, id.size)) for i in range(id.size)]

    def VectorConstant(self, vec):
        return [self(NumberConstant(v)) for v in tolist(vec.value)]

    def MatrixConstant(self, mtx):
        return [[self(NumberConstant(v)) for v in tolist(row)] for row in mtx.value]

    def MakeVector(self, vec):
        return [self(el) for el in vec.elems]

    def IfThenElse(self, node):
        """Implementation for IfThenElse for 'listwise' translators.
        Relies on the existence of self._iif function."""
        
        test, iftrue, iffalse = self(node.test), self(node.iftrue), self(node.iffalse)
        if isinstance(iftrue, list):
            if not isinstance(iffalse, list) or len(iftrue) != len(iffalse):
                raise ValueError("Mixed types in IfThenElse expressions are not supported")
            return [self._iif(test, ift, iff) for ift, iff in zip(iftrue, iffalse)]
        else:
            if isinstance(iffalse, list):
                raise ValueError("Mixed types in IfThenElse expressions are not supported")
            return self._iif(test, iftrue, iffalse)
    
    def LFold(self, node, **kw):
        "If we know vectors are lists, we can improve LFold to avoid deep recursions"

        if not node.elems:
            raise ValueError("LFold expects at least one element")
        op = self(node.op, **kw)
        return reduce(lambda x, y: _apply_bin_op(node.op, op, x, y),
                      [self(el, **kw) for el in node.elems])

    VecSum = is_(lambda vec: reduce(lambda x, y: x + y, vec))


class StandardArithmetics:
    """A partial implementation of an AST processor,
    which assimes that:
        - all binary and unary operations are interpreted as lambda functions.
        - basic arithmetics and comparisons map to Python's basic arithmetics.
        - sigmoid can be expressed in terms of Exp as usual.
    """

    Mul = is_(lambda x, y: x * y)
    Div = is_(lambda x, y: x / y)
    Add = is_(lambda x, y: x + y)
    Sub = is_(lambda x, y: x - y)
    USub = is_(lambda x: -x)
    LtEq = is_(lambda x, y: x <= y)
    Eq = is_(lambda x, y: x == y)

    def Sigmoid(self, _):
        return lambda x: 1/(1 + self(Exp())(-x))


def prepare_assign_to(assign_to, n_actual_targets):
    """Converts the value of the assign_to parameter to a list of strings, as needed.
    
    >>> prepare_assign_to('x', 1)
    ['x']
    >>> prepare_assign_to('x', 2)
    ['x1', 'x2']
    >>> prepare_assign_to(['x'], 1)
    ['x']
    >>> prepare_assign_to(['a','b'], 2)
    ['a', 'b']
    >>> prepare_assign_to(None, 3)
    >>> prepare_assign_to(['a'], 2)
    Traceback (most recent call last):
    ...
    ValueError: The number of outputs (2) does not match the number of assign_to values (1)
    """

    if assign_to is None:
        return None

    if isinstance(assign_to, str):
        if n_actual_targets == 1:
            return [assign_to]
        else:
            return ['{0}{1}'.format(assign_to, i+1) for i in range(n_actual_targets)]
    
    if len(assign_to) != n_actual_targets:
        raise ValueError(("The number of outputs ({0}) does not match the number"
                          " of assign_to values ({1})").format(n_actual_targets, len(assign_to)))
    
    return assign_to


def id_generator(template='_tmp{0}', start=1):
    return map(template.format, count(start))


def denumpyfy(value):
    if hasattr(value, 'dtype'):
        return value.item()
    else:
        return value



"""
A set of convenience wrapper functions for AST creation.
"""
#pylint: disable=protected-access
import numpy as np
from . import ast

def const(value):
    ## Convenience function for creating Number, Vector or MatrixConstants.
    if np.isscalar(value):
        return ast.NumberConstant(value)
    elif hasattr(value, 'ndim'):
        if value.ndim == 0:
            return ast.NumberConstant(value.item())
        if value.ndim == 1:
            return ast.VectorConstant(value)
        elif value.ndim == 2:
            return ast.MatrixConstant(value)
        else:
            raise ValueError("Only one or two-dimensional vectors are supported")
    elif hasattr(value, '__iter__') or hasattr(value, '__next__'):
        return const(np.asarray(value))
    else:
        raise ValueError("Invalid constant: {0}".format(value))

class _FuncCreator:
    ## An object similar to sqlalchemy.func
    def __getattr__(self, attrname):
        if attrname != '__wrapped__': # Otherwise test discovery crashes
            return getattr(ast, attrname, None)()

func = _FuncCreator()

def vector(elems):
    if hasattr(elems, '__next__'):
        elems = list(elems)
    return ast.MakeVector(elems)

def ident(name, size=None):
    return ast.VectorIdentifier(name, size) if size else ast.Identifier(name)

def ref(name, to_obj=None):
    if to_obj is not None:
        try:
            size = len(to_obj)
        except ast.UnableToDecompose():
            size = None
        return ast.TypedReference(name, to_obj._dtype, size)
    else:
        return ast.Reference(name)

def defn(**kw):
    for name, value in kw.items():
        return ast.Definition(name, value)

def let(*steps):
    return ast.Let(list(steps[:-1]), steps[-1])

def iif(test, iftrue, iffalse):
    return ast.IfThenElse(test, iftrue, iffalse)

def repeat(node, n_times):
    return ast.MakeVector([node]*n_times)

def mean(elems, vector_dim=None):
    divisor = len(elems)
    if vector_dim is not None:
        divisor = [divisor] * vector_dim
    return sum_(elems) / const(divisor)

def sum_(elems):
    return ast.LFold(func.Add, elems)



"""
Commonly useful functions.
"""
from skompiler.dsl import ident, vector


def is_(x):
    return lambda self, node, **kw: x

def prepare_inputs(inputs, n_features=None):
    if hasattr(inputs, '__next__'):
        # Unroll iterators
        inputs = [next(inputs) for i in range(n_features)]
    if isinstance(inputs, str):
        if not n_features:
            raise ValueError("Impossible to determine number of input variables")
        return ident(inputs, size=n_features)
    elif isinstance(inputs, list):
        if n_features is not None and len(inputs) != n_features:
            raise ValueError("The number of inputs must match the number of features in the tree")
        if isinstance(inputs[0], str):
            inputs = [ident(el) for el in inputs]
        return vector(inputs)
    else:
        return inputs



"""
Common useful functionality.
"""
from skompiler.dsl import func, let, defn, ref, repeat


def classifier(probs, method):
    """Given a probability expression and a method name, returns the classifier output"""

    if method == 'predict_proba':
        return probs
    elif method == 'predict':
        return func.ArgMax(probs)
    elif method == 'predict_log_proba':
        return func.Log(probs)
    else:
        raise ValueError("Invalid method: {0}".format(method))

def vecsumnormalize(node, vector_dim):
    return let(defn(x=node),
               defn(s=func.VecSum(ref('x'))),
               ref('x') / repeat(ref('s'), vector_dim))

def sklearn_softmax(node, vector_dim):
    return let(defn(x=node),
               defn(xmax=func.VecMax(ref('x'))),
               defn(xfix=ref('x') - repeat(ref('xmax'), vector_dim)),
               func.Softmax(ref('xfix')))



"""
SKLearn model transformation to SKompiler's AST.
"""
#pylint: disable=unused-argument
try:
    from functools import singledispatch
except ImportError:
    # Python 3.3-
    from singledispatch import singledispatch
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model.base import LinearModel
from sklearn.svm import SVC, SVR
from sklearn.tree.tree import BaseDecisionTree, DecisionTreeRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor,\
                             GradientBoostingClassifier, GradientBoostingRegressor,\
                             AdaBoostClassifier
from sklearn.cluster import KMeans
from sklearn.decomposition.pca import _BasePCA
from sklearn.neural_network import MLPClassifier, MLPRegressor

from sklearn.preprocessing import Binarizer, MinMaxScaler, MaxAbsScaler, StandardScaler,\
                                  Normalizer
from sklearn.pipeline import Pipeline

from skompiler.dsl import ident, vector

from .._common import prepare_inputs
from .linear_model.logistic import logreg_binary, logreg_multiclass
from .linear_model.base import linear_model
from .tree.base import decision_tree
from .ensemble.forest import random_forest_classifier, random_forest_regressor
from .ensemble.gradient_boosting import gradient_boosting_classifier, gradient_boosting_regressor
from .ensemble.weight_boosting import adaboost_classifier
from .cluster.k_means import k_means
from .decomposition.pca import pca
from .neural_network.multilayer_perceptron import mlp, mlp_classifier
from .preprocessing.data import binarize, scale, unscale, standard_scaler, normalizer


@singledispatch
def translate(model, inputs='x', method='predict'):
    """
    Translate a given SKLearn model to a SK AST expression.

    Kwargs:

      inputs (string or ASTNode):
            The name of the variable that will be used to represent
            the input in the resulting expression. It can be given either as a
            single string (in this case the variable denotes an input vector),
            a list of strings (in this case each element denotes the name of one input component of a vector),
            or an ASTNode (e.g. VectorIdentifier, or vector([ident('x'), ident('y')]))
        
      method (string):  Method to be expressed. Possible options:
            'predict', for all supported models.
            'decision_function', for all supported models.
            'predict_proba', 'predict_log_proba': for classifiers.
    """
    raise NotImplementedError("Conversion not implemented for {0}".format(model.__class__.__name__))

_supported_methods = {}

# An improved version of @translate.register decorator
def register(cls, methods):
    def decorator(fn):
        @translate.register(cls)
        def new_fn(model, inputs='x', method='predict'):
            if method not in methods:
                raise ValueError("Method {0} is not supported (or not implemented yet) for {1}".format(method, cls.__name__))
            return fn(model, inputs, method)
        return new_fn
    _supported_methods[cls] = methods
    return decorator


@register(LogisticRegression, ['decision_function', 'predict', 'predict_proba', 'predict_log_proba'])
def _(model, inputs, method):
    ovr = (model.multi_class in ["ovr", "warn"] or
           (model.multi_class == 'auto' and (model.classes_.size <= 2 or
                                             model.solver == 'liblinear')))
    if model.coef_.shape[0] == 1: # Binary logreg
        if not ovr:
            raise NotImplementedError("Logistic regression with binary outcomes and multinomial outputs is not implemented")
            # ... It's not too hard, actually, just need to find the 15 minutes needed to implement it.
        return logreg_binary(model.coef_.ravel(), model.intercept_[0], inputs=prepare_inputs(inputs, model.coef_.shape[-1]), method=method)
    else: # Multiclass logreg
        return logreg_multiclass(model.coef_, model.intercept_, method=method,
                                 inputs=prepare_inputs(inputs, model.coef_.shape[-1]), multi_class='ovr' if ovr else 'multinomial')

@register(SVC, ['decision_function', 'predict'])
def _(model, inputs, method):
    # For linear SVC the predict and decision function logic is the same as for logreg
    if model.kernel != 'linear':
        raise NotImplementedError("Translation for nonlinear SVC not implemented")
    if model.decision_function_shape != 'ovr':
        raise NotImplementedError("Translation not implemented for one-vs-one SVC")
    if len(model.classes_) > 2 and method != 'predict':
        raise NotImplementedError("Translation not implemented for non-binary SVC") # See sklearn.utils.multiclass._ovr_decision_function
    if model.coef_.shape[0] == 1: # Binary
        return logreg_binary(model.coef_.ravel(), model.intercept_[0], inputs=prepare_inputs(inputs, model.coef_.shape[-1]), method=method)
    else: # Multiclass logreg
        return logreg_multiclass(model.coef_, model.intercept_, method=method,
                                 inputs=prepare_inputs(inputs, model.coef_.shape[-1]), multi_class='ovr')

@register(SVR, ['predict'])
def _(model, inputs, method):
    if isinstance(model, SVR) and model.kernel != 'linear':
        raise NotImplementedError("Nonlinear SVR not implemented")
    return linear_model(model.coef_.ravel(), model.intercept_.item(), prepare_inputs(inputs, model.coef_.shape[-1]))

@register(LinearModel, ['predict'])
def _(model, inputs, method):
    return linear_model(model.coef_.ravel(), model.intercept_, prepare_inputs(inputs, model.coef_.shape[-1]))
    
@register(BaseDecisionTree, ['predict', 'predict_proba', 'predict_log_proba'])
def _(model, inputs, method):
    if isinstance(model, DecisionTreeRegressor) and method != 'predict':
        raise ValueError("Method {0} is not supported for DecisionTreeRegressor".format(method))
    return decision_tree(model.tree_, prepare_inputs(inputs, model.n_features_), method)

@register(RandomForestClassifier, ['predict', 'predict_proba', 'predict_log_proba'])
def _(model, inputs, method):
    return random_forest_classifier(model, prepare_inputs(inputs, model.n_features_), method)

@register(RandomForestRegressor, ['predict'])
def _(model, inputs, method):
    return random_forest_regressor(model, prepare_inputs(inputs, model.n_features_))

@register(GradientBoostingClassifier, ['decision_function'])
def _(model, inputs, method):
    return gradient_boosting_classifier(model, prepare_inputs(inputs, model.n_features_))

@register(GradientBoostingRegressor, ['predict'])
def _(model, inputs, method):
    return gradient_boosting_regressor(model, prepare_inputs(inputs, model.n_features_))

@register(AdaBoostClassifier, ['decision_function', 'predict', 'predict_proba', 'predict_log_proba'])
def _(model, inputs, method):
    return adaboost_classifier(model, prepare_inputs(inputs, model.estimators_[0].n_features_), method)

@register(KMeans, ['transform', 'predict'])
def _(model, inputs, method):
    return k_means(model.cluster_centers_, prepare_inputs(inputs, model.cluster_centers_.shape[1]), method)

@register(_BasePCA, ['transform'])
def _(model, inputs, method):
    return pca(model, prepare_inputs(inputs, model.components_.shape[1]))

@register(Pipeline, ['predict', 'predict_proba', 'decision_function', 'predict_log_proba', 'transform'])
def _(model, inputs, method):
    if not model.steps:
        raise ValueError("Empty pipeline provided")
    # The first step in the pipeline is responsible for preparing the inputs
    first_method = 'transform' if len(model.steps) > 1 else method
    expr = translate(model.steps[0][1], inputs, first_method)
    for i in range(1, len(model.steps)-1):
        expr = translate(model.steps[i][1], expr, 'transform')
    if len(model.steps) > 1:
        expr = translate(model.steps[-1][1], expr, method)
    return expr

@register(MLPRegressor, ['predict'])
def _(model, inputs, method):
    return mlp(model, prepare_inputs(inputs, len(model.coefs_[0])))

@register(MLPClassifier, ['predict', 'predict_proba', 'predict_log_proba'])
def _(model, inputs, method):
    return mlp_classifier(model, prepare_inputs(inputs, len(model.coefs_[0])), method)

@register(Binarizer, ['transform'])
def _(model, inputs, method):
    return binarize(model.threshold, prepare_inputs(inputs))

@register(MinMaxScaler, ['transform'])
def _(model, inputs, method):
    return scale(model.scale_, model.min_, prepare_inputs(inputs, len(model.scale_)))

@register(MaxAbsScaler, ['transform'])
def _(model, inputs, method):
    return unscale(model.scale_, prepare_inputs(inputs, len(model.scale_)))

@register(StandardScaler, ['transform'])
def _(model, inputs, method):
    n = None
    if model.with_mean:
        n = len(model.mean_)
    elif model.with_std:
        n = len(model.scale_)
    return standard_scaler(model, prepare_inputs(inputs, n))

@register(Normalizer, ['transform'])
def _(model, inputs, method):
    return normalizer(model.norm, prepare_inputs(inputs))



"""
SKompiler: Generate Python AST expressions from SKAST.
Useful for testing and evaluating expressions.

Note that the expressions with a Let condition are compiled differently to
those without.

>>> from ..toskast import string

Bare expressions without Let compile to bare Python expressions.
You need to wrap them in ast.Expression to use and may eval to get a value:

>>> expr = string.translate("12.4 * (b[1] + Y[0])")
>>> pyast = translate(expr)
>>> code = compile(ast.Expression(body=pyast), "__main__", "eval")
>>> eval(code, {'b': [10, 20, 30], 'Y': [1.2]})
262.88

Let-expressions compile to ast.Module with multiple expressions, the last of which
writes the result value to the __result__ global. You may only exec this code:

>>> expr = string.translate("a=X[1]; b=a+1; 12.4 * (b + Y[0])")
>>> pyast = translate(expr)
>>> code = compile(pyast, "__main__", "exec")
>>> vars = {}
>>> eval(code, {'X': [10, 20, 30], 'Y': [1.2]}, vars)
>>> vars['__result__']
275.28

In general, use utils.evaluate to evaluate SKAST expressions.
"""
import ast
import numpy as np
from ..ast import USub, Identifier, NumberConstant, IsBoolean, merge_let_scopes, Max
from ._common import ASTProcessor, StandardOps, denumpyfy


_linearg = dict(lineno=1, col_offset=0) # Most Python AST nodes require these

def translate(node, dialect=None):
    """
    When dialect is None, translates the given SK AST expression to a Python AST tree.
    Otherwise, further converts the tree depending on the value of dialect, to:
    
      'code':   Python source code (via the astor package)
      'lambda': Executable function (via expr.lambdify)
    
    >>> from skompiler.toskast.string import translate as skast
    >>> expr = skast('[2*x[0], 1] if x[1] <= 3 else [12.0, 45.5]')
    >>> print(translate(expr, 'code'))
    (np.array([2 * x[0], 1]) if x[1] <= 3 else np.array([12.0, 45.5]))
    <BLANKLINE>
    >>> fn = translate(expr, 'lambda')
    >>> fn(x=[1, 2])
    array([2, 1])
    """
    pyast = PythonASTWriter()(merge_let_scopes(node))
    if dialect is None:
        return pyast
    elif dialect == 'lambda':
        return lambdify(pyast)
    elif dialect == 'code':
        import astor
        code = astor.to_source(pyast)
        # Replace some internal identifiers with matching functions
        code = code.replace('__np__', 'np').replace('__exp__', 'np.exp').replace('__log__', 'np.log')
        code = code.replace('__argmax__', 'np.argmax').replace('__sum__', 'np.sum')
        return code
    else:
        raise ValueError("Unknown dialect: {0}".format(dialect))

def _ident(name):
    "Shorthand for defining methods in PythonASTWriter (see code below)"
    return lambda self, x: self(Identifier(name))

def _is(node):
    "Shorthand for defining methods in PythonASTWriter (see code below)"
    return lambda self, x: node

class PythonASTWriter(ASTProcessor, StandardOps):
    """
    An AST processor, which translates a given SKAST node to a Python AST.

    >>> import ast
    >>> topy = PythonASTWriter()
    >>> print(ast.dump(topy(Identifier('x'))))
    Name(id='x', ctx=Load())
    """

    def Identifier(self, name):
        return ast.Name(id=name.id, ctx=ast.Load(), **_linearg)
    VectorIdentifier = Identifier

    def IndexedIdentifier(self, sub):
        return ast.Subscript(value=self(Identifier(sub.id)),
                             slice=ast.Index(value=self(NumberConstant(sub.index))),
                             ctx=ast.Load(), **_linearg)

    def NumberConstant(self, num):
        return ast.Num(n=denumpyfy(num.value), **_linearg)

    def VectorConstant(self, vec):
        result = ast.parse('__np__.array()', mode='eval').body
        result.args = [ast.List(elts=[ast.Num(n=denumpyfy(el), **_linearg) for el in vec.value],
                                ctx=ast.Load(), **_linearg)]
        return result

    def MakeVector(self, mv):
        result = ast.parse('__np__.array()', mode='eval').body
        result.args = [ast.List(elts=[self(el) for el in mv.elems],
                                ctx=ast.Load(), **_linearg)]
        return result

    def MatrixConstant(self, mat):
        result = ast.parse('__np__.array()', mode='eval').body
        result.args = [ast.List(elts=[ast.List(elts=[ast.Num(n=denumpyfy(el), **_linearg) for el in row],
                                               ctx=ast.Load(), **_linearg) for row in mat.value], ctx=ast.Load(), **_linearg)]
        return result

    def UnaryFunc(self, node, **kw):
        if isinstance(node.op, USub):
            return ast.UnaryOp(op=self(node.op), operand=self(node.arg), **_linearg)
        else:
            return ast.Call(func=self(node.op), args=[self(node.arg)], keywords=[], **_linearg)

    def BinOp(self, node, **kw):
        op, left, right = self(node.op), self(node.left), self(node.right)
        if isinstance(node.op, IsBoolean):
            return ast.Compare(left=left, ops=[op], comparators=[right], **_linearg)
        elif isinstance(node.op, Max):
            return ast.Call(func=self(node.op), args=[self(node.left), self(node.right)], keywords=[], **_linearg)
        else:
            return ast.BinOp(op=op, left=left, right=right, **_linearg)
    
    def IfThenElse(self, node):
        return ast.IfExp(test=self(node.test), body=self(node.iftrue), orelse=self(node.iffalse), **_linearg)
    
    def Let(self, node, **kw):
        code = [ast.Assign(targets=[ast.Name(id='_def_' + defn.name, ctx=ast.Store(), **_linearg)],
                           value=self(defn.body), **_linearg) for defn in node.defs]
        # Evaluate the expression body into a "__result__" variable
        code.append(
            ast.Assign(targets=[ast.Name(id='__result__', ctx=ast.Store(), **_linearg)],
                       value=self(node.body), **_linearg))
        return ast.Module(body=code)

    def Reference(self, ref):
        return ast.Name(id='_def_' + ref.name, ctx=ast.Load(), **_linearg)
    TypedReference = Reference

    # Functions
    Exp = _ident('__exp__')
    Sqrt = _ident('__sqrt__')
    Log = _ident('__log__')
    Step = _ident('__step__')
    VecSum = _ident('__sum__')
    ArgMax = _ident('__argmax__')
    Sigmoid = _ident('__sigmoid__')
    Softmax = _ident('__softmax__')
    VecMax = _ident('__vecmax__')
    Max = _ident('__max__')
    Abs = _ident('__abs__')

    # Operators
    Mul = _is(ast.Mult())
    Div = _is(ast.Div())
    Add = _is(ast.Add())
    Sub = _is(ast.Sub())
    USub = _is(ast.USub())
    DotProduct = _is(ast.MatMult())
    MatVecProduct = DotProduct

    # Predicates
    LtEq = _is(ast.LtE())
    Eq = _is(ast.Eq())

# ------------- Evaluation of Python AST-s --------------- #

def _softmax(X):
    X = np.exp(X)
    sum_prob = np.sum(X, axis=1).reshape((-1, 1))
    X /= sum_prob
    return X

_eval_vars = {
    '__np__': np,
    '__exp__': np.exp,
    '__sqrt__': np.sqrt,
    '__log__': np.log,
    '__sum__': np.sum,
    '__argmax__': np.argmax,
    '__vecmax__': np.max,
    '__max__': np.maximum,
    '__abs__': np.abs,
    '__sigmoid__': lambda z: 1.0/(1.0 + np.exp(-z)),
    '__sum_normalize__': lambda x: x / np.sum(x),
    '__softmax__': lambda x: _softmax([x])[0, :],
    '__step__': lambda x: 1 if x > 0 else 0  # This is how step is implemented in LogisticRegression
}

def lambdify(pyast):
    """
    Converts a given Python AST, produced by PythonASTWriter to an executable Python function.

    >>> from ..ast import NumberConstant, BinOp, Mul, Identifier
    >>> pyast = translate(BinOp(Mul(), NumberConstant(2), Identifier('x')))
    >>> fn = lambdify(pyast)
    >>> fn(x=3.14)
    6.28
    """
    if isinstance(pyast, ast.Module):
        # Exec the code:
        code = compile(pyast, "__main__", "exec")
        def result(**inputs):
            globals_ = {}
            globals_.update(_eval_vars)
            eval(code, inputs, globals_)  # pylint: disable=eval-used
            return globals_['__result__']
    else:
        # Eval the code:
        code = compile(ast.Expression(body=pyast), "__main__", "eval")
        def result(**inputs):
            globals_ = {}
            globals_.update(_eval_vars)
            return eval(code, inputs, globals_)  # pylint: disable=eval-used
    return result



"""
Translation for Theano graphs (limited to those generatable via Keras default operations)
"""
#pylint: disable=protected-access
from collections import OrderedDict
from skompiler import ast
from skompiler.dsl import const, func, vector, ref
from .._common import is_


def _dot(x, y):
    xtype = x._dtype
    ytype = y._dtype
    if xtype == ast.DTYPE_VECTOR and ytype == ast.DTYPE_VECTOR:
        return x @ y
    elif xtype == ast.DTYPE_VECTOR and ytype == ast.DTYPE_MATRIX:
        # Keras multiplies matrices on the right, we reverse this
        return y @ x
    else:
        raise NotImplementedError("Dot product not implemented for given arguments")

def _auto_broadcasting(elemwise_func):
    # The theano graph actually contains explicit instructions
    # on how the values must be broadcast. However, at the moment we do not carry
    # this information with us during translation and simply assume that the only
    # kind of broadcasting possible is turning a constant into a vector.
    def fn(x, y=None):
        if y is not None:
            xtype, ytype = x._dtype, y._dtype
            if xtype == ast.DTYPE_VECTOR and ytype == ast.DTYPE_SCALAR:
                y = vector([y]*len(x))
            elif xtype == ast.DTYPE_SCALAR and ytype == ast.DTYPE_VECTOR:
                x = vector([x]*len(y))
            return elemwise_func(x, y)
        else:
            return elemwise_func(x)
    return fn

_elemwise_ops = {
    'Add': func.Add,
    'Mul': func.Mul,
    'Abs': func.Abs,
    'Tanh': lambda x: func.Sigmoid(x*2)*2 - 1,
    'ScalarSigmoid': func.Sigmoid,
    'TrueDiv': func.Div,
    'Cast': lambda x: x, # Ignore casts for now
    'LT': func.LtEq,     # TODO: Support for LessThan
}


class TheanoTranslator:
    def __init__(self, input_var_name, inputs):
        self.input_var_name = input_var_name
        self.inputs = inputs
        self.definitions = OrderedDict()
    
    def __call__(self, node, **kw):
        cname = node.__class__.__name__
        return getattr(self, cname)(node, **kw)
    
    def TensorVariable(self, node):
        if node.name == self.input_var_name:
            return self.inputs
        elif node.name == 'keras_learning_phase':
            return False
        elif node.owner is None:
            raise NotImplementedError("Do not know how to handle tensor variable {0}".format(node.name))
        else:
            return self(node.owner)
    
    def TensorSharedVariable(self, node):
        if node.name is None:
            return const(node.container.value)
        refname = node.name.replace('/', '_')
        if refname not in self.definitions:
            res = const(node.container.value)
            if isinstance(res, ast.MatrixConstant):
                # Keras always multiplies matrices on the right, we prefer the other way
                res = const(res.value.T)
            self.definitions[refname] = res
        return ref(refname, self.definitions[refname])

    def TensorConstant(self, node):
        return const(node.data)

    def Apply(self, node):
        # Special case to support Dropout without the need to parse the training phase branch
        if node.op.__class__.__name__ == 'IfElse':
            test = self(node.inputs[0])
            if isinstance(test, bool):
                return self(node.inputs[1]) if test else self(node.inputs[2])
            else:
                return ast.IfThenElse(test, self(node.inputs[1]), self(node.inputs[2]))
        op = self(node.op)
        inputs = [self(inp) for inp in node.inputs]
        return op(*inputs)

    def DimShuffle(self, node):
        o = node.new_order
        if o not in [('x', 'x'), ('x', 0)]:
            raise NotImplementedError("Unusual DimShuffle not implemented")
        # If o is ('x', 'x'), the argument is a constant that must be broadcasted
        # If o is ('x', 0) the argument is a vector that is taken as-i
        return lambda x: x

    def Elemwise(self, node):
        opname = node.scalar_op.__class__.__name__
        op = _elemwise_ops.get(opname, None)
        if not op:
            raise NotImplementedError("Elemwise operation {0} not implemented".format(opname))
        return _auto_broadcasting(op)

    Dot = is_(_dot)
    Softmax = is_(func.Softmax)



"""
PCA implementation
"""
import numpy as np
from skompiler.dsl import const

def pca(model, inputs):
    matrix = np.array(model.components_)
    if model.whiten:
        matrix /= np.sqrt(model.explained_variance_)[:, np.newaxis]
    if model.mean_ is not None:
        inputs = inputs - const(model.mean_)
    return const(matrix) @ inputs



from skompiler.ast import decompose
from skompiler.dsl import const, vector, iif, let, defn, func, ref

def binarize(threshold, inputs):
    if not isinstance(inputs, list):
        inputs = decompose(inputs)
    return vector([iif(inp <= const(threshold), const(0), const(1)) for inp in inputs])

def scale(scale_, min_, inputs):
    return inputs * const(scale_) + const(min_)

def unscale(scale_, inputs):
    return inputs / const(scale_)

def standard_scaler(model, inputs):
    if model.with_mean:
        inputs = inputs - const(model.mean_)
    if model.with_std:
        inputs = inputs / const(model.scale_)
    return inputs

def normalizer(norm, inputs):
    if norm == 'l2':
        norm = func.Sqrt(func.VecSum(inputs * inputs))
    elif norm == 'l1':
        norm = func.VecSum(func.Abs(inputs))
    elif norm == 'max':
        norm = func.VecMax(inputs)
    else:
        raise ValueError("Unknown norm {0}".format(norm))
    return let(defn(norm=norm),
               defn(norm_fix=iif(norm == const(0), const(1), norm)),
               inputs / vector([ref('norm_fix')]*len(inputs)))



"""
A convenience interface to SKompiler's functionality.
Wraps around the intricacies of the various toskast/fromskast pieces.
"""


def skompile(*args, inputs=None):
    """
    Creates a SKAST expression from a given bound method of a fitted SKLearn model.
    A shorthand notation for SKompiledModel(method, inputs)

    Args:

        args:  Either a bound method of a trained model (e.g. skompile(model.predict_proba)),
               OR two arguments - a model and a method name (e.g. skompile(model, 'predict_proba')
               (which may be necessary for some models where the first option cannot be used due to metaclasses)
        
        inputs:  A string or a list of strings, or a SKAST node or a list of SKAST nodes,
                 denoting the input variable(s) to your model.
                 A single string corresponds to a vector variable (which will be indexed to access
                 the components). A list of strings corresponds to a vector with separately named components.
               You may pass the inputs as a non-keyword argument as well (the last one in *args)
               If not specified, the default value of 'x' is used.
    
    Returns:
        An instance of SKompiledModel.

    Examples:
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.linear_model import LogisticRegression
        >>> X, y = load_iris(True)
        >>> m = LogisticRegression().fit(X, y)
        >>> print(skompile(m.predict))
        argmax((([[...]] m@v x) + [ ...]))
        >>> print(skompile(m, 'predict'))
        argmax((([[...]] m@v x) + [ ...]))
        >>> print(skompile(m.predict, 'y'))
        argmax((([[...]] m@v y) + [ ...]))
        >>> print(skompile(m.predict, ['x','y','z','w']))
        argmax((([[...]] m@v [x, y, z, w]) + [ ...]))
        >>> print(skompile(m, 'predict', 'y'))
        argmax((([[...]] m@v y) + [ ...]))
        >>> print(skompile(m, 'predict', ['x','y','z','w']))
        argmax((([[...]] m@v [x, y, z, w]) + [ ...]))
        >>> from skompiler.ast import VectorIdentifier, Identifier
        >>> print(skompile(m, 'predict', VectorIdentifier('y', 4)))
        argmax((([[...]] m@v y) + [ ...]))
        >>> print(skompile(m, 'predict', map(Identifier, ['x','y','z','w'])))
        argmax((([[...]] m@v [x, y, z, w]) + [ ...]))
        >>> from sklearn.pipeline import Pipeline
        >>> p = Pipeline([('1', m)])
        >>> skompile(p.predict)
        Traceback (most recent call last):
        ...
        ValueError: The bound method ... Please, use the skompile(m, 'predict') syntax instead.
    """

    if len(args) > 3:
        raise ValueError("Too many arguments")
    elif not args:
        raise ValueError("Invalid arguments")
    elif len(args) == 3:
        if inputs is not None:
            raise ValueError("Too many arguments")
        model, method, inputs = args
    elif len(args) == 2:
        if hasattr(args[0], '__call__'):
            model, method = _get_model_and_method(args[0])
            inputs = args[1]
        else:
            model, method = args
    else:
        model, method = _get_model_and_method(args[0])
    if not inputs:
        inputs = 'x'
    return _translate(model, inputs, method)

def _translate(model, inputs, method):
    if model.__class__.__module__.startswith('keras.'):
        if method != 'predict':
            raise ValueError("Only the 'predict' method is supported for Keras models")
        # Import here, this way we do not force everyone to install everything
        from .toskast.keras import translate as from_keras
        return from_keras(model, inputs)
    else:
        from .toskast.sklearn import translate as from_sklearn
        return from_sklearn(model, inputs=inputs, method=method)

def _get_model_and_method(obj):
    if not hasattr(obj, '__call__'):
        raise ValueError("Please, provide a method to compile.")
    if not hasattr(obj, '__self__'):
        raise ValueError("The bound method object was probably mangled by "
                         "SKLearn's metaclasses and cannot be passed to skompile as skompile(m.predict). "
                         "Please, use the skompile(m, 'predict') syntax instead.")
    return obj.__self__, obj.__name__



"""
SKLearn logistic regression to SKAST.
"""
from skompiler.dsl import func
from .base import linear_model
from ..common import classifier, vecsumnormalize, sklearn_softmax


def logreg_binary(coef, intercept, inputs, method="predict_proba"):
    """
    Binary logistic regression.

    Args:

       inputs:  a list of AST nodes to be used as inputs to the model.

    Kwargs:

        method (string): The sklearn method's output to emulate.
            'decision_function' - The logistic regression decision function only.
            'predict_proba' - The output will be the probability of class 1
                    (note that it is NOT two probabilities, as in case of SKLearn's
                    actual predict_proba output)
            'predict_log_proba' - The log probability of class 1
            'predict'  - The output will be an integer 0/1, predicting the class.
    """
    decision = linear_model(coef, intercept, inputs)

    if method == "decision_function":
        return decision
    if method == "predict":
        return func.Step(decision)

    return classifier(func.Sigmoid(decision), method)

    
def logreg_multiclass(coef_matrix, intercept_vector, inputs='x', method="predict_proba", multi_class='ovr'):
    """
    Multiclass logistic regression.

    Kwargs:

      output (string): The sklearn method's output to emulate.
          'decision_function' - The logistic regression decision function only.
          'predict_proba' - The output will be the probability of class 1
                (note that it is NOT two probabilities, as in case of SKLearn's
                 actual predict_proba output)
          'predict_log_proba' - The log probability of class 1
          'predict'  - The output will be an integer 0/1, predicting the class.
      
      inputs (string or ASTNode): The name of the inputs variable to use in the formula.
                    Any SKAST expression could be used instead, assuming it encodes an
                    input vector.
    
      multi_class (string):  The value of the "multi_class" setting used in the model. Either 'ovr' or 'multinomial'
    """
    decision = linear_model(coef_matrix, intercept_vector, inputs)

    if method == "decision_function":
        return decision
    if method == "predict":
        return func.ArgMax(decision)

    if multi_class == 'ovr':
        probs = vecsumnormalize(func.Sigmoid(decision), coef_matrix.shape[0])
    elif multi_class == 'multinomial':
        probs = sklearn_softmax(decision, coef_matrix.shape[0])
    else:
        raise ValueError("Invalid value of the multi_class argument: " + multi_class)

    return classifier(probs, method)
    


"""
Decision trees to SKAST
"""
from skompiler.dsl import sum_
from ..common import classifier
from ..tree.base import decision_tree


def random_forest_classifier(model, inputs, method="predict_proba"):
    """
    Creates a SKAST expression corresponding to a given random forest classifier
    """
    trees = [decision_tree(estimator.tree_, inputs, method="predict_proba", value_transform=lambda v: v/len(model.estimators_))
             for estimator in model.estimators_]
    return classifier(sum_(trees), method)


def random_forest_regressor(model, inputs):
    """
    Creates a SKAST expression corresponding to a given random forest regressor
    """

    return sum_([decision_tree(estimator.tree_, inputs=inputs, method="predict", value_transform=lambda v: v/len(model.estimators_))
                 for estimator in model.estimators_])



"""
String to SKAST translator.
"""
import ast
from .python import translate as from_python


def translate(expr):
    """Convert a given (restricted) Python code string to a SK-AST.
       
       So far we only need this functionality for debugging purposes,
       hence instead of implementing a full-fledged parser, we rely on
       Python's ast.parse.

       This means that if the expression contains non-Python code
       or Python code which we cannot translate, you will get cryptic errors.

       >>> expr = translate("12.4 * (X1[25.3] + Y)")
       >>> print(str(expr))
       (12.4 * (X1[25.3] + Y))
       >>> expr = translate("a=X; b=a+2; 12.4 * (a[25.3] + b + y)")
       Traceback (most recent call last):
       ...
       ValueError: Subscripting named references is not supported
       >>> expr = translate("a=X; b=a+2; 12.4 * (a + b + y)")
       >>> print(str(expr))
       {
       $a = X;
       $b = ($a + 2);
       (12.4 * (($a + $b) + y))
       }
       """
    return from_python(ast.parse(expr))



"""
Decision trees to SKAST
"""
import numpy as np
from skompiler.ast import decompose
from skompiler.dsl import const, iif


def decision_tree(tree, inputs, method="predict", value_transform=None):
    """
    Creates a SKAST expression corresponding to a given SKLearn Tree object.

    Kwargs:
      
       inputs:  a list of AST nodes to be used as inputs to the model.
       method:  'predict' (for classifier and regressor models),
                'predict_proba' or 'predict_log_proba' (for classifier models)
        
       value_transform:  If not None, the tree values are processed using the given operator.
                This way we may propagate constant operations into trees
                (e.g. instead of const * decision_tree(...) you may just have a
                 decision tree which outputs const * value)
    
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.tree import DecisionTreeClassifier
    >>> from skompiler.dsl import ident, vector
    >>> m = DecisionTreeClassifier(max_depth=2, random_state=1).fit(*load_iris(True))
    >>> print(decision_tree(m.tree_, ident('x', m.n_features_)))
    (if (x[3] <= 0.80...) then 0 else (if (x[3] <= 1.75) then 1 else 2))

    >>> inputs = vector(map(ident, 'abcd'))
    >>> print(decision_tree(m.tree_, inputs, method='predict_proba'))
    (if (d <= 0.80...) then [1. 0. 0.] else (if (d <= 1.75) then [0... 0.90... 0.09...] else [0... 0.02... 0.97...]))
    """
    v = tree.value[:, 0, :]
    if v.shape[1] == 1:
        # Regression model
        if method != 'predict':
            raise ValueError("Only predict method is supported for regression trees")
        v = v[:, 0]
    else:
        # Classifier
        if method == "predict":
            v = np.argmax(v, axis=1)
        else:
            v = v / v.sum(axis=1)[:, np.newaxis]
            if method == "predict_log_proba":
                v = np.log(v)
            elif method != "predict_proba":
                raise ValueError("Invalid method: {0}".format(method))

    if value_transform:
        v = value_transform(v)
    return TreeWalker(tree, inputs, v).walk()


class TreeWalker:
    """Converts a SKLearn Tree object to a SKAST expression.
    
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.tree import DecisionTreeRegressor
    >>> from skompiler.dsl import ident
    >>> m =  DecisionTreeRegressor(max_depth=2, random_state=1).fit(*load_iris(True))
    >>> tr = TreeWalker(m.tree_, ident('x', 4))
    >>> print(tr.walk())
    (if (x[3] <= 0.80...) then 0.0 else (if (x[3] <= 1.75) then 1.09... else 1.97...))
    >>> tr = TreeWalker(m.tree_, ident('x', 4), np.arange(m.tree_.node_count))
    >>> print(tr.walk())
    (if (x[3] <= 0.80...) then 1 else (if (x[3] <= 1.75) then 3 else 4))
    """
    
    def __init__(self, tree, features, node_values=None):
        """
        Kwargs:
           node_values (list/array): A way to override the tree.value array.
                                     Must be a 1D or 2D array of values.
        """
        if not isinstance(features, list):
            features = decompose(features)
        
        self.tree = tree
        self.features = features
        if tree.n_features != len(features):
            raise ValueError("Incorrect number of features provided")
        if node_values is None:
            self.values = tree.value[:, 0]
            if self.values.shape[1] == 1:
                self.values = self.values[:, 0]
        else:
            self.values = node_values
        
    def walk(self, node_id=0):
        if node_id >= self.tree.node_count or node_id < 0:
            raise ValueError("Invalid node id")
        if self.tree.children_left[node_id] == -1:
            if self.tree.children_right[node_id] != -1:
                raise ValueError("Invalid tree structure. Children must either be both present or absent")
            
            if self.values.ndim == 1:
                return const(self.values[node_id].item())
            else:
                return const(self.values[node_id])
        else:
            ft = self.tree.feature[node_id]
            if ft < 0 or ft >= self.tree.n_features:
                raise ValueError("Invalid feature value for node {0}".format(node_id))
            return iif(self.features[ft] <= const(self.tree.threshold[node_id].item()),
                       self.walk(self.tree.children_left[node_id]),
                       self.walk(self.tree.children_right[node_id]))



"""
SKompiler: Generate Sympy expressions from SKAST.
"""
import numpy as np
import sympy as sp
from ..ast import IsElemwise, Mul
from ._common import ASTProcessor, is_, StandardOps, StandardArithmetics


def translate(node, dialect=None, true_argmax=True, assign_to='y', component=None, lambdify_inputs_str='x', **kw):
    """Translates SKAST to a Sympy expression and optionally generates code from it.

    KwArgs:
        dialect (string):   If None, returns the Sympy expression. Otherwise translates it further to one of the supported languages.
                            Supported values:
                                'c', 'cxx', 'rust', 'fortran', 'js', 'r', 'julia', 'mathematica', 'octave', 'lambda'.
                            If dialect == 'lambda', the expression is lambdified (the call is then equivalent to
                            lambdify(lambdify_inputs_str, translate(node)))
        
        true_argmax (bool): When True (default), the generated expression will include a "Sympy-executable" definition of argmax(vector)
                            (in the form of a lengthy expression "if v[0] == max(v) then 0, else if v[1] == max(v) then 1, ...")
                            When False, the expression will contain just the name "argmax". This may be sufficient
                            if you only need Sympy as an intermediate representation before compiling into a different language.

        assign_to:          This value is passed further to sympy code printers when dialect is not None.
                            When assign_to is not None, it specifies that the code printer should generate assignment statements,
                            setting values of the given variable. Otherwise, a pure expression is generated (which is not always
                            possible if the expression is a matrix)
        
        component (int):    If the result is a vector, return only the specified component of it.

        lambdify_inputs_str (str): The string specifying lambdified function inputs (when dialect == 'lambda')

        **kw:               Other arguments passed to the code generator (when dialect is not None)
    
    >>> from skompiler.toskast.string import translate as skast
    >>> expr = skast('[2*x[0], 1] if x[1] <= 3 else [12.0, 45.5]')
    >>> print(translate(expr, 'js'))
    if (x_1 <= 3) {
       y[0] = 2*x_0;
    }
    else {
       y[0] = 12.0;
    }
    if (x_1 <= 3) {
       y[1] = 1;
    }
    else {
       y[1] = 45.5;
    }
    >>> print(translate(expr, 'js', assign_to=None, component=0))
    ((x_1 <= 3) ? (
       2*x_0
    )
    : (
       12.0
    ))
    """
    syexpr = SympyWriter(true_argmax=true_argmax)(node)
    if component is not None:
        syexpr = syexpr[component]
    if dialect is None:
        return syexpr
    elif dialect == 'lambda':
        return lambdify(lambdify_inputs_str, syexpr)
    else:
        return to_code(syexpr, dialect, assign_to=assign_to, **kw)

def _argmax(val):
    "A sympy implementation of argmax"

    maxval = sp.Max(*val)
    pieces = [(i, sp.Eq(val[i], maxval)) for i in range(len(val)-1)]
    pieces.append((len(val)-1, True))
    return sp.Piecewise(*pieces)

def _max(val):
    return sp.Max(*val)

def _softmax(vec):
    "A sympy implementation of sklearn's softmax"
    sexp = [sp.exp(vec[i]) for i in range(len(vec))]
    return sp.ImmutableMatrix([sexp[i] / sum(sexp) for i in range(len(sexp))])


class SympyWriter(ASTProcessor, StandardOps, StandardArithmetics):
    """A SK AST processor, producing a Sympy expression"""

    def __init__(self, true_argmax=False):
        self.true_argmax = true_argmax
    
    def Identifier(self, id):
        return sp.symbols(id.id)

    def VectorIdentifier(self, node):
        # This is not the best option, because this way all our vector inputs
        # must be 2D matrices (for purposes of lambdify as well as printingc:w
        x = sp.MatrixSymbol(node.id, node.size, 1)
        return sp.ImmutableMatrix([x[i] for i in range(node.size)])

        # NB: This alone won't work, because Dot operator wants to have an actual matrix as input
        # return sp.MatrixSymbol(node.id, node.size, 1)
        #
        # This (and  the version with x = sp.IndexedBase('x')) is not good because it breaks code printers
        # (apparently they expect all indices to have lower and upper bounds which does not happen if you provide
        #  numeric indices, or smth like that, probably a bug somewhere in sympy)
        # x = sp.MatrixSymbol(node.id, node.size, 1)
        # return sp.ImmutableMatrix([x[i] for i in range(node.size)])

    def IndexedIdentifier(self, sub):
        if sub.size is None:  # Hack to handle Python-written expressions. We do not represent them as "true" indexed values in Sympy
            return sp.symbols('{0}_{1}'.format(sub.id, sub.index))
        else:
            return sp.IndexedBase(sub.id, shape=(sub.size,))[sub.index]

    def NumberConstant(self, num):
        return sp.sympify(num.value)

    def VectorConstant(self, vec):
        return sp.ImmutableMatrix(vec.value)

    MatrixConstant = VectorConstant

    def MakeVector(self, vec):
        return sp.ImmutableMatrix([self(el) for el in vec.elems])

    def UnaryFunc(self, node, **kw):
        arg = self(node.arg)
        op = self(node.op)
        if isinstance(arg, sp.MatrixBase) and isinstance(node.op, IsElemwise):
            if arg.shape[1] != 1:
                raise NotImplementedError("Elementwise operations are only supported for vectors (column matrices)")
            return sp.ImmutableMatrix([op(arg[i]) for i in range(len(arg))])
        else:
            return op(arg)

    def BinOp(self, node, **kw):
        left = self(node.left)
        right = self(node.right)
        op = self(node.op)
        if isinstance(left, sp.MatrixBase) and isinstance(right, sp.MatrixBase) and isinstance(node.op, IsElemwise):
            if left.shape != right.shape:
                raise ValueError("Shapes of the arguments do not match")
            if left.shape[1] != 1:
                raise NotImplementedError("Elementwise operations are only supported for vectors (column matrices)")
            return sp.ImmutableMatrix([op(left[i], right[i]) for i in range(len(left))])
        else:
            return op(left, right)
    
    def IfThenElse(self, node):
        # Piecewise function with matrix output is not a Matrix itself, which breaks some of the logic
        # Hence this won't work in general:
        test, iftrue, iffalse = self(node.test), self(node.iftrue), self(node.iffalse)
        if hasattr(iftrue, 'shape'):
            if iftrue.shape != iffalse.shape:
                raise ValueError("Shapes of the IF branches must match")
            if iftrue.shape[1] != 1:
                raise NotImplementedError("Elementwise operations are only supported for vectors (column matrices)")
            return sp.ImmutableMatrix([sp.Piecewise((ift, test), (iff, True)) for ift, iff in zip(iftrue, iffalse)])
        else:
            return sp.Piecewise((iftrue, test), (iffalse, True))

    def MatVecProduct(self, _):
        return self(Mul())
    
    DotProduct = is_(lambda x, y: x.dot(y))
    Exp = is_(sp.exp)
    Log = is_(sp.log)
    Step = is_(sp.Heaviside)
    Sqrt = is_(sp.sqrt)
    Abs = is_(sp.Abs)
    VecSum = is_(sum) # Yes, we return a Python summation here
    Softmax = is_(_softmax)
    VecMax = is_(_max)
    Max = is_(lambda x, y: _max([x, y]))
    Eq = is_(sp.Eq)

    def ArgMax(self, _):
        return _argmax if self.true_argmax else sp.Function('argmax')

# Utility function

def _softmax(X):
    X = np.exp(X)
    sum_prob = np.sum(X, axis=1).reshape((-1, 1))
    X /= sum_prob
    return X

_lambdify_modules = ["numpy",
                     {"Heaviside": lambda x: int(x > 0),
                      "softmax": lambda x: _softmax([x])[0, :]
                     }]

def lambdify(sympy_inputs_str, sympy_expr):
    return sp.lambdify(sp.symbols(sympy_inputs_str), sympy_expr, modules=_lambdify_modules)



_ufns = {'argmax': 'argmax'}

#pylint: disable=unnecessary-lambda
_code_printers = {
    'c': lambda expr, **kw: sp.ccode(expr, standard='c99', user_functions=_ufns, **kw),
    'cxx': lambda expr, **kw: sp.cxxcode(expr, user_functions=_ufns, **kw),
    'rust': lambda expr, **kw: sp.rust_code(expr, user_functions=_ufns, **kw),
    'fortran': lambda expr, **kw: sp.fcode(expr, standard=95, user_functions=_ufns, **kw),
    'js': lambda expr, **kw: sp.jscode(expr, user_functions=_ufns, **kw),
    'r': lambda expr, **kw: sp.rcode(expr, user_functions=_ufns, **kw),
    'julia': lambda expr, **kw: sp.julia_code(expr, user_functions=_ufns, **kw),
    'mathematica': lambda expr, assign_to=None, **kw: sp.mathematica_code(expr, user_functions=_ufns, **kw),
    'octave': lambda expr, **kw: sp.octave_code(expr, user_functions=_ufns, **kw),
}

def to_code(syexpr, dialect, assign_to='y', **kw):
    '''
    Shorthand for converting the resulting Sympy expression to code in various languages.


    Args:

        dialect (str): The target language. Can be one of:
            'c', 'cxx', 'rust', 'fortran', 'js', 'r', 'julia', 'mathematica', 'octave'.

    Kwargs:

        assign_to (str or sympy expression):
            Passed to the sympy code printers.
            If this is not None, the generated code is a set of assignments to the target variable
            (or, of the result is an array, to the components of the target array).
            Note that if assign_to is None, and the resulting expression is a vector (e.g. predict_proba vector
            of probabilities), sympy may be unable to generate meaningful code for many languages,
            because those lack the "Matrix" datatype (hence it cannot be the result of a one-liner expression)
    '''
    if dialect not in _code_printers:
        raise ValueError("Unknown dialect: ")
    return _code_printers[dialect](syexpr, assign_to=assign_to, **kw)



"""
K-means implementation.
"""
from skompiler.dsl import const, func, let, defn, ref, vector


def k_means(cluster_centers, inputs, method):
    sq_dists = vector(let(defn(dx=inputs - const(c)),
                          ref('dx') @ ref('dx'))
                      for c in cluster_centers)
    if method == 'transform':
        return func.Sqrt(sq_dists)
    elif method == 'predict':
        return func.ArgMax(sq_dists * -1)
    else:
        raise ValueError("Unsupported methods for KMeans: {0}".format(method))



"""
SKLearn linear model to SKAST.
"""
from skompiler.dsl import const

def linear_model(coef, intercept, inputs):
    """
    Linear regression.
    Depending on the shape of the coef and intercept, produces either a single-valued
    linear model (w @ x + b) or a multi-valued one (M @ x + b_vec)

    Args:

        coef (np.array): A vector (1D array, for single-valued model) or a matrix (2D array, for multi-valued one) for the model.
        intercept:  a number (for single-valued) or a 1D array (for multi-valued regression).
        inputs:  a list of AST nodes to be used as the input vector to the model or a single node, corresponding to a vector.
    """

    single_valued = (coef.ndim == 1)
    if single_valued and hasattr(intercept, '__iter__'):
        raise ValueError("Single-valued linear model must have a single value for the intercept")
    elif not single_valued and (coef.ndim != 2 or intercept.ndim != 1):
        raise ValueError("Multi-valued linear model must have a 2D coefficient matrix and a 1D intercept vector")

    return const(coef) @ inputs + const(intercept)



'''
Python AST to SKAST translator.

May only convert nodes of Python AST which have a mapping in SK-AST.
Useful for debugging testing and simplistic parsing mostly:


>>> expr = ast.parse("12.4 * (X1[25.3] + Y)")
>>> print(str(translate(expr)))
(12.4 * (X1[25.3] + Y))
>>> expr = ast.parse("a=12+b; b=2*a; 12.4 * (X[25.3] + Y + 2*a*b)")
>>> print(str(translate(expr)))
{
$a = (12 + b);
$b = (2 * $a);
(12.4 * ((X[25.3] + Y) + ((2 * $a) * $b)))
}
'''
#pylint: disable=wildcard-import,unused-wildcard-import,unused-argument
import ast
from skompiler.ast import *
from ._common import is_

def translate(node):
    return PythonASTProcessor()(node)

_funcmap = {
    'log': Log(),
    'exp': Exp(),
    'step': Step(),
    'sqrt': Sqrt(),
    'abs': Abs(),
}

class PythonASTProcessor:

    def __call__(self, node, **kw):
        cls = node.__class__.__name__
        if not hasattr(self, cls):
            raise NotImplementedError("No translation logic implemented for node type {0}".format(cls))
        return getattr(self, cls)(node, **kw)

    def Module(self, module, local_varnames=None):
        # Module may have more than one statement. We only allow a sequence of Assign statements,
        # followed by an expression
        definitions = []
        local_varnames = local_varnames or set()
        for assn in module.body[:-1]:
            if not isinstance(assn, ast.Assign):
                raise NotImplementedError("Only a sequence of assignments followed by an expression is allowed")
            if len(assn.targets) != 1:
                raise NotImplementedError("Assignment to a single variable allowed only")
            if not isinstance(assn.targets[0], ast.Name):
                raise NotImplementedError("Assignment may only be done to a named variable")
            varname = assn.targets[0].id
            definitions.append(Definition(varname, self(assn.value, local_varnames=local_varnames)))
            local_varnames.add(varname)

        body = self(module.body[-1], local_varnames=local_varnames)
        if local_varnames:
            return Let(definitions, body)
        return body

    def Expr(self, expr, **kw):
        return self(expr.value, **kw)

    def Expression(self, expr, **kw):
        return self(expr.body, **kw)

    def Name(self, name, local_varnames=None):
        if local_varnames and name.id in local_varnames:
            return Reference(name.id)
        else:
            return Identifier(name.id)

    def Subscript(self, sub, local_varnames=None):
        if not isinstance(sub.slice, ast.Index) or not isinstance(sub.value, ast.Name) \
           or not isinstance(sub.slice.value, ast.Num):
            raise NotImplementedError("Unsupported form of subscript")
        if local_varnames and sub.value.id in local_varnames:
            raise ValueError("Subscripting named references is not supported")
        return IndexedIdentifier(id=sub.value.id,
                                 index=sub.slice.value.n,
                                 size=None) # This makes Sympy sad

    def Num(self, num, **kw):
        return NumberConstant(num.n)

    def UnaryOp(self, op, **kw):
        return UnaryFunc(op=self(op.op, **kw),
                         arg=self(op.operand, **kw))

    def BinOp(self, op, **kw):
        return BinOp(op=self(op.op, **kw),
                     left=self(op.left, **kw),
                     right=self(op.right, **kw))
    
    def Call(self, call, **kw):
        if not isinstance(call.func, ast.Name) or call.keywords or len(call.args) != 1:
            raise ValueError("Only one-argument functions are supported")
        if call.func.id not in _funcmap:
            raise ValueError("Unsupported unary function: " + call.func.id)
        return UnaryFunc(op=_funcmap[call.func.id], arg=self(call.args[0], **kw))

    def IfExp(self, ifexp, **kw):
        return IfThenElse(self(ifexp.test, **kw), self(ifexp.body, **kw), self(ifexp.orelse, **kw))
    
    def Compare(self, cmp, **kw):
        if len(cmp.comparators) != 1 or len(cmp.ops) != 1:
            raise ValueError("Only one-element comparison expressions are supported")
        return BinOp(self(cmp.ops[0], **kw), self(cmp.left, **kw), self(cmp.comparators[0], **kw))
    
    def List(self, lst, **kw):
        return MakeVector([self(el, **kw) for el in lst.elts])
    
    Mult = is_(Mul())
    Add = is_(Add())
    Sub = is_(Sub())
    USub = is_(USub())
    LtE = is_(LtEq())
    Div = is_(Div())



"""
Decision trees to SKAST
"""
from skompiler.dsl import vector, const, sum_
from ..tree.base import decision_tree


def gradient_boosting_classifier(model, inputs, method="decision_function"):
    """
    Creates a SKAST expression corresponding to a given gradient boosting classifier

    At the moment we only support model's decision_function method.
    FYI: Conversion to probabilities and a prediction depends on the loss and by default
          is done as np.exp(score - (logsumexp(score, axis=1)[:, np.newaxis])))
    """

    if method != "decision_function":
        raise NotImplementedError("Only decision_function is implemented for gradient boosting models so far")

    tree_exprs = [vector([decision_tree(estimator.tree_, inputs, method="predict", value_transform=lambda v: v * model.learning_rate)
                          for estimator in iteration])
                  for iteration in model.estimators_]
    return sum_(tree_exprs + [const(model.init_.priors)])


def gradient_boosting_regressor(model, inputs, method="decision_function"):
    """
    Creates a SKAST expression corresponding to a given GB regressor.
    
    The logic is mostly the same as for the classifier, except we work with scalars rather than vectors.
    """

    if method != "decision_function":
        raise NotImplementedError("Only decision_function is implemented for gradient boosting models so far")
    
    tree_exprs = [decision_tree(iteration[0].tree_, inputs, method="predict", value_transform=lambda v: v * model.learning_rate)
                  for iteration in model.estimators_]
    return sum_(tree_exprs + [const(model.init_.mean)])



'''
SKompiler: Library for converting trained SKLearn models into abstract expressions suitable
for further compilation into executable code in various languages.

Author: Konstantin Tretyakov
License: MIT
'''
from .api import skompile

__version__ = '0.5.2'



'''
SKompiler: Library for converting trained SKLearn models into abstract expressions suitable
for further compilation into executable code in various languages.

Author: Konstantin Tretyakov
License: MIT
'''

from setuptools import setup, find_packages

extra_requires = []
try:
    from functools import singledispatch #pylint: disable=unused-import
except ImportError:
    # Python 3.3-
    extra_requires.append("singledispatch")

setup(name='SKompiler',
      version=[ln for ln in open("skompiler/__init__.py") if ln.startswith("__version__")][0].split("'")[1],
      description="Library for compiling trained SKLearn models into abstract expressions "
                  "suitable for further compilation into executable code in various languages.",
      long_description=open("README.md", encoding='utf-8').read(),
      long_description_content_type="text/markdown",
      classifiers=[ # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Code Generators',
          'License :: OSI Approved :: MIT License',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
      ],
      keywords='sklearn datascience modelling deployment',
      author='Konstantin Tretyakov',
      author_email='kt@ut.ee',
      url='https://github.com/konstantint/SKompiler',
      license='MIT',
      packages=find_packages(exclude=["examples", "tests"]),
      include_package_data=True,
      zip_safe=True,
      install_requires=["scikit-learn"] + extra_requires,
      extras_require={
          "dev": ["sympy", "sqlalchemy", "pylint", "pytest", "jupyter", "twine", "pandas",
                  "keras", "theano"],
          "full": ["sympy", "sqlalchemy", "astor >= 0.6"]
      }
     )



"""
SKompiler: Generate SQLAlchemy expressions from SKAST.
"""
from functools import reduce
from collections import namedtuple
import numpy as np
import sqlalchemy as sa
from sqlalchemy.sql.selectable import Join, FromGrouping
from ..ast import ArgMax, VecMax, Softmax, IsElemwise, VecSum, Max, IsAtom
from ._common import ASTProcessor, StandardOps, StandardArithmetics, is_, tolist,\
                     not_implemented, prepare_assign_to, id_generator, denumpyfy

#pylint: disable=trailing-whitespace
def translate(node, dialect=None, assign_to='y', component=None,
              multistage=True, key_column='id', from_obj='data'):
    """Translates SKAST to an SQLAlchemy expression (or a list of those, if the output should be a vector).

    If dialect is not None, further compiles the expression(s) to a given dialect via to_sql.
    
    Kwargs:
       assign_to (None/string/list of str): See to_sql
       
       component (int):    If the result is a vector and you only need one component of it, specify its index (0-based) here.

       multistage (bool):  When multistage=False, the returned value is a single expression which can be selected directly from the
                           source data table. This, however, may make the resulting query rather long, as some functions (e.g. argmax)
                           require repeated computaion of the same parts over and over.
                           The problem is solved by splitting the computation in a sequence of CTE subqueries - the "multistage" mode.
                           The resulting query may then look like
                           
                           with _tmp1 as (select [probability computations] from data),
                                _tmp2 as (select [argmax computation] from _tmp1),
                                ... 
                           select [final values] from _tmpX
                        
                           Default - True

       from_obj:    A string or a SQLAlchemy selectable object - the source table for the data.
                    In non-multistage mode this may be None - in this case the returned value is 
                    simply 'SELECT cols'.

       key_column:  A string or a sa.column object, naming the key column in the source table.
                    Compulsory for multistage mode.

    
    >>> from skompiler.toskast.string import translate as skast
    >>> expr = skast('[2*x[0], 1] if x[1] <= 3 else [12.0, 45.5]')
    >>> print(translate(expr, 'sqlite', multistage=False, from_obj=None))
    SELECT CASE WHEN (x2 <= 3) THEN 2 * x1 ELSE 12.0 END AS y1, CASE WHEN (x2 <= 3) THEN 1 ELSE 45.5 END AS y2


    >>> expr = skast('x=1; y=2; x+y')
    >>> print(translate(expr, 'sqlite', multistage=True))
    WITH _tmp1 AS 
    (SELECT data.id AS __id__, 1 AS f1 
    FROM data), 
    _tmp2 AS 
    (SELECT data.id AS __id__, 2 AS f1 
    FROM data)
     SELECT _tmp1.f1 + _tmp2.f1 AS y 
    FROM _tmp1 JOIN _tmp2 ON _tmp1.__id__ = _tmp2.__id__
    >>> expr = skast('x+y')
    >>> stbl = sa.select([sa.column('id'), sa.column('x'), sa.column('y')], from_obj=sa.table('test')).cte('_data')
    >>> print(translate(expr, 'sqlite', multistage=False, from_obj=stbl))
    WITH _data AS 
    (SELECT id, x, y 
    FROM test)
     SELECT x + y AS y 
    FROM _data
    """
    if multistage and from_obj is None:
        raise ValueError("from_obj must be specified in multistage mode")
    result = SQLAlchemyWriter(from_obj=from_obj, key_column=key_column, multistage=multistage)(node)

    if component is not None:
        result = result._replace(cols=[result.cols[component]])
    
    assign_to = prepare_assign_to(assign_to, len(result.cols))
    if assign_to is not None:
        result = result._replace(cols=[col.label(lbl) for col, lbl in zip(result.cols, assign_to)])

    result = sa.select(result.cols, from_obj=result.from_obj)

    if dialect is not None:
        result = to_sql(result, dialect)
    return result


def _max(xs):
    if len(xs) == 1:
        return xs[0]
    return reduce(greatest, xs)

def _sum(iterable):
    "The built-in 'sum' does not work for us as we need."
    return reduce(lambda x, y: x+y, iterable)

def _iif(cond, iftrue, iffalse):
    # Optimize if (...) then X else X for literal X
    # A lot of these occur when compiling trees
    if isinstance(iftrue, sa.sql.elements.BindParameter) and \
       isinstance(iffalse, sa.sql.elements.BindParameter) and \
       iftrue.value == iffalse.value:
        return iftrue
    return sa.case([(cond, iftrue)], else_=iffalse)

def _matvecproduct(M, x):
    return [_sum(m_i[j] * x[j] for j in range(len(x))) for m_i in M]

def _dotproduct(xs, ys):
    return [_sum(x * y for x, y in zip(xs, ys))]

def _step(x):
    return _iif(x > 0, 1, 0)

def extract_tables(from_obj):
    if isinstance(from_obj, FromGrouping):
        return extract_tables(from_obj.element)
    elif isinstance(from_obj, Join):
        return extract_tables(from_obj.left) + extract_tables(from_obj.right)
    else:
        return [from_obj]

def _merge(tbl1, tbl2):
    if tbl1 is None:
        return tbl2
    elif tbl2 is None:
        return tbl1
    if tbl1 is tbl2:
        return tbl1
    # Either of the arguments may be a join clause and these
    # may include repeated elements. If so, we have to extract them and recombine.
    all_tables = list(sorted(set(extract_tables(tbl1) + extract_tables(tbl2)), key=lambda x: x.name))
    tbl1 = all_tables[0]
    joined = tbl1
    for tbl_next in all_tables[1:]:
        joined = joined.join(tbl_next, onclause=tbl1.key_ == tbl_next.key_)
        joined.key_ = tbl1.key_
    return joined

Result = namedtuple('Result', 'cols from_obj')

class SQLAlchemyWriter(ASTProcessor, StandardOps, StandardArithmetics):
    """A SK AST processor, producing a SQLAlchemy "multistage" expression.
       The interpretation of each node is a tuple, containing a list of column expressions and a from_obj,
       where these columns must be queried from."""
    
    def __init__(self, from_obj='data', key_column='id',
                 positive_infinity=float(np.finfo('float64').max),
                 negative_infinity=float(np.finfo('float64').min),
                 multistage=True):
        self.positive_infinity = positive_infinity
        self.negative_infinity = negative_infinity
        if multistage:
            if isinstance(from_obj, str):
                from_obj = sa.table(from_obj, sa.column(key_column))
                # This is a bit hackish, but quite convenient.
                # This way we do not have to carry around an extra "key" field in our results all the time
                from_obj.key_ = from_obj.columns[key_column]
            else:
                if key_column not in from_obj:
                    raise ValueError("The provided selectable does not contain the key column {0}".format(key_column))
                from_obj.key_ = from_obj.columns[key_column]
        elif isinstance(from_obj, str):
            from_obj = sa.table(from_obj)
        self.from_obj = from_obj
        self.temp_ids = id_generator()
        self.references = [{}]
        self.multistage = multistage
    
    def Identifier(self, id):
        return Result([sa.column(id.id)], self.from_obj)

    def _indexed_identifier(self, id, idx):
        return sa.column("{0}{1}".format(id, idx+1))
    
    def IndexedIdentifier(self, sub):
        return Result([self._indexed_identifier(sub.id, sub.index)], self.from_obj)

    def _number_constant(self, value):
        # Infinities have to be handled separately
        if np.isinf(value):
            value = self.positive_infinity if value > 0 else self.negative_infinity
        else:
            value = denumpyfy(value)
        return sa.literal(value)

    def NumberConstant(self, num):
        return Result([self._number_constant(num.value)], self.from_obj)

    def VectorIdentifier(self, id):
        return Result([self._indexed_identifier(id.id, i) for i in range(id.size)], self.from_obj)

    def VectorConstant(self, vec):
        return Result([self._number_constant(v) for v in tolist(vec.value)], self.from_obj)

    def MatrixConstant(self, mtx):
        return Result([[self._number_constant(v) for v in tolist(row)] for row in mtx.value], self.from_obj)

    def UnaryFunc(self, node, **kw):
        arg = self(node.arg)
        if isinstance(node.op, ArgMax):
            return self._argmax(arg)
        elif isinstance(node.op, VecMax):
            return self._vecmax(arg)
        elif isinstance(node.op, VecSum):
            return self._vecsum(arg)
        elif isinstance(node.op, Softmax):
            return self._softmax(arg)
        else:
            op = self(node.op)
            return Result([op(el) for el in arg.cols], arg.from_obj)

    ArgMax = VecSumNormalize = VecSum = VecMax = Softmax = not_implemented

    def BinOp(self, node, **kw):
        left, right, op = self(node.left), self(node.right), self(node.op)
        if not isinstance(node.op, IsElemwise):
            # MatVecProduct requires atomizing the argument, otherwise it will be repeated multiple times in the output
            if not isinstance(node.right, IsAtom):
                right = self._make_cte(right)
            return Result(op(left.cols, right.cols), _merge(left.from_obj, right.from_obj))
        elif len(left.cols) != len(right.cols):
            raise ValueError("Mismatching operand dimensions in {0}".format(repr(node.op)))
        elif isinstance(node.op, Max):
            # Max is implemented as (if x > y then x else y), hence to avoid double-computation,
            # we save x and y in separate CTE's
            if not isinstance(node.left, IsAtom):
                left = self._make_cte(left)
            if not isinstance(node.right, IsAtom):
                right = self._make_cte(right)
            return Result([op(lc, rc) for lc, rc in zip(left.cols, right.cols)], _merge(left.from_obj, right.from_obj))
        else:
            return Result([op(lc, rc) for lc, rc in zip(left.cols, right.cols)], _merge(left.from_obj, right.from_obj))

    def MakeVector(self, vec):
        result = []
        tbls = set()
        for el in vec.elems:
            el = self(el)
            tbls.add(el.from_obj)
            if len(el.cols) != 1:
                raise ValueError("MakeVector expects a list of scalars")
            result.append(el.cols[0])
        tbls = list(tbls)
        target_table = tbls[0]
        for tbl in tbls[1:]:
            new_joined = target_table.join(tbl, onclause=target_table.key_ == tbl.key_)
            new_joined.key_ = target_table.key_
            target_table = new_joined
        return Result(result, target_table)

    def IfThenElse(self, node):
        test, iftrue, iffalse = self(node.test), self(node.iftrue), self(node.iffalse)

        return Result([_iif(test.cols[0], ift, iff) for ift, iff in zip(iftrue.cols, iffalse.cols)],
                      reduce(_merge, [test.from_obj, iftrue.from_obj, iffalse.from_obj]))

    MatVecProduct = is_(_matvecproduct)
    DotProduct = is_(_dotproduct)
    Exp = is_(sa.func.exp)
    Log = is_(sa.func.log)
    Sqrt = is_(sa.func.sqrt)
    Abs = is_(sa.func.abs)
    Step = is_(_step)
    Max = is_(lambda x, y: _max([x, y]))

    # ------ The actual "multi-stage" logic -----
    def Let(self, node, **kw):
        if not self.multistage:
            return StandardOps.Let(self, node, **kw)
        self.references.append({})
        for defn in node.defs:
            self.references[-1][defn.name] = self._make_cte(self(defn.body))
        result = self(node.body)
        self.references.pop()
        return result

    def Reference(self, node):
        if not self.multistage:
            raise ValueError("References are not supported in non-multistage mode")
        if node.name not in self.references[-1]:
            raise ValueError("Undefined reference: {0}".format(node.name))
        return self.references[-1][node.name]

    def _make_cte(self, result, col_names=None, key_label='__id__'):
        if not self.multistage:
            return result
        if col_names is None:
            col_names = ['f{0}'.format(i+1) for i in range(len(result.cols))]
        labeled_cols = [c.label(n) for c, n in zip(result.cols, col_names)]
        new_tbl = sa.select([result.from_obj.key_.label(key_label)] + labeled_cols, from_obj=result.from_obj).cte(next(self.temp_ids))
        new_tbl.key_ = new_tbl.columns[key_label]
        new_cols = [new_tbl.columns[n] for n in col_names]
        return Result(new_cols, new_tbl)

    def _argmax(self, result):
        if len(result.cols) == 1:
            return Result([sa.literal(0)], self.from_obj)
        features = self._make_cte(result)
        max_val = Result([_max(features.cols)], features.from_obj)
        max_val = self._make_cte(max_val, ['_max'])

        argmax = sa.case([(col == max_val.cols[0], i)
                          for i, col in enumerate(features.cols[:-1])],
                         else_=len(features.cols)-1)
        return Result([argmax], _merge(features.from_obj, max_val.from_obj))
    
    def _vecmax(self, result):
        return Result([_max(result.cols)], result.from_obj)
    
    def _softmax(self, result):
        return self._vecsumnormalize(Result([sa.func.exp(col) for col in result.cols], result.from_obj))

    def _vecsumnormalize(self, result):
        features = self._make_cte(result)
        sum_val = Result([_sum(features.cols)], features.from_obj)
        sum_val = self._make_cte(sum_val, ['_sum'])
        return Result([col/sum_val.cols[0] for col in features.cols],
                      _merge(features.from_obj, sum_val.from_obj))

    def _vecsum(self, result):
        return Result([_sum(result.cols)], result.from_obj)

# ------- SQLAlchemy "greatest" function
# See https://docs.sqlalchemy.org/en/latest/core/compiler.html
#pylint: disable=wrong-import-position,wrong-import-order
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import Numeric

class greatest(expression.FunctionElement):
    type = Numeric()
    name = 'greatest'

@compiles(greatest)
def default_greatest(element, compiler, **kw):
    res = compiler.visit_function(element, **kw)
    return res

@compiles(greatest, 'sqlite')
@compiles(greatest, 'mssql')
@compiles(greatest, 'oracle')
def case_greatest(element, compiler, **kw):
    arg1, arg2 = list(element.clauses)
    return compiler.process(sa.case([(arg1 > arg2, arg1)], else_=arg2), **kw)


# Utilities ----------------------------------
import sqlalchemy.dialects
#pylint: disable=wildcard-import,unused-wildcard-import
from sqlalchemy.dialects import *   # Must do it in order to getattr(sqlalchemy.dialects, ...)
def to_sql(sa_expr, dialect_name='sqlite'):
    """
    Helper function. Given a SQLAlchemy expression, returns the corresponding
    SQL string in a given dialect.
    """
    
    dialect_module = getattr(sqlalchemy.dialects, dialect_name)
    return str(sa_expr.compile(dialect=dialect_module.dialect(),
                               compile_kwargs={'literal_binds': True}))



"""
Multilayer perceptron
"""
from skompiler.dsl import func, const
from ..common import classifier

_activations = {
    'identity': lambda x: x,
    'tanh': lambda x: func.Sigmoid(x*2)*2 - 1,
    'logistic': func.Sigmoid,
    'relu': lambda x: func.Max(x, const([0] * len(x))),
    'softmax': func.Softmax
}

def mlp(model, inputs):
    actns = [model.activation]*(len(model.coefs_)-1) + [model.out_activation_]
    outs = inputs
    for M, b, a in zip(model.coefs_, model.intercepts_, actns):
        outs = _activations[a](const(M.T) @ outs + const(b))
    return outs

def mlp_classifier(model, inputs, method):
    out = mlp(model, inputs)
    if model.n_outputs_ == 1 and method == 'predict':
        # Binary classifier
        return func.Step(out - 0.5)
    else:
        return classifier(out, method)



"""
Translation from Keras.
So far just a proof of concept, working via Theano.
"""
from keras.models import Sequential
from skompiler.ast import Let, Definition, inline_definitions
from .._common import prepare_inputs
from .theano import TheanoTranslator


def translate(model, inputs='x'):
    if not isinstance(model, Sequential):
        raise NotImplementedError("Support for non-sequential models is not implemented")

    if len(model.inputs) != 1:
        raise NotImplementedError("Support for models with multiple inputs is not implemented")
    
    if len(model.outputs) != 1:
        raise NotImplementedError("Support for models with multiple outputs is not implemented")

    if not model.outputs[0].__class__.__module__.startswith('theano.'):
        raise NotImplementedError("Non-theano backend support is not implemented")

    if len(model.input_shape) != 2:
        raise NotImplementedError("Support for models with non-vector inputs is not implemented")
    
    if model.input_shape[0] is not None:
        raise NotImplementedError("Not sure what to do when model.input_shape[0] is not None")
    
    inputs = prepare_inputs(inputs, model.input_shape[1])
    tt = TheanoTranslator(model.inputs[0].name, inputs)
    result = tt(model.outputs[0])
    if tt.definitions:
        result = Let([Definition(name, value) for name, value in tt.definitions.items()],
                     result)
        # For now excel/sql code generators cannot handle reference variables which point to matrix constants.
        # which TheanoTranslator generates.
        # We overcome the problem by dumb inlining
        result = inline_definitions(result)
    return result
